from immuneML.dsl.symbol_table.SymbolType import SymbolType


class SymbolTableEntry:

    def __init__(self, symbol: str, symbol_type: SymbolType, item, config: dict = None):
        self.symbol = symbol
        self.symbol_type = symbol_type
        self.item = item
        self.config = config



from multiprocessing.pool import Pool

import numpy as np
from editdistance import eval as edit_distance

from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.data_model.SequenceSet import ReceptorSequence
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.encodings.reference_encoding.SequenceMatchingSummaryType import SequenceMatchingSummaryType


class SequenceMatcher:
    """
    Matches the sequences across the given list of reference sequences (a list of ReceptorSequence objects) and returns the following information:
    {
        "repertoires":[{
            "sequences": [{
                "sequence": "AAA",
                "matching_sequences": ["AAA", "AAC"],
                "v_gene": "V12",
                "j_gene": "J3",
                "locus": "A"
            }], # list of sequences for the repertoire with matched sequences for each original sequence
            "repertoire": "fdjshfk321231", # repertoire identifier
            "repertoire_index": 2,  # the index of the repertoire in the dataset,
            "sequences_matched": 4,  # number of sequences from the repertoire which are a match for at least one reference sequence
            "percentage_of_sequences_matched": 0.75,  # percentage of sequences from the repertoire that have at least one match in the reference sequences
            "metadata": {"CD": True},  # dict with parameters that can be used for analysis on repertoire level and that serve as a starting point for label configurations
            "locus": ["A","B"] # list of chains in the repertoire
        }, ...]
    }
    """

    CORES = 4

    def match(self, dataset: RepertoireDataset, reference_sequences: list, max_distance: int, summary_type: SequenceMatchingSummaryType) -> dict:

        matched = {"repertoires": []}

        for index, repertoire in enumerate(dataset.get_data()):
            matched["repertoires"].append(self.match_repertoire(repertoire, index,
                                                                reference_sequences, max_distance, summary_type))

        return matched

    def matches_gene(self, gene1, gene2):
        """for two genes (e.g. "TRBV12-2,TRBV12-4" and "TRBV12-2") returns True if they share at least one gene
        (here TRBV12-2)"""
        genes1 = [g.split("*")[0] for g in gene1.split(",")] if gene1 not in ['', None] else []
        genes2 = [g.split("*")[0] for g in gene2.split(",")] if gene2 not in ['', None] else []

        return len(set(genes1).intersection(set(genes2))) > 0 or (len(genes1) == len(genes2) and len(genes1) == 0)

    def matches_sequence(self, original_sequence: ReceptorSequence, reference_sequence: ReceptorSequence, max_distance):
        """
        :param original_sequence: ReceptorSequence
        :param reference_sequence: ReceptorSequence
        :param max_distance: max allowed Levenshtein distance between two sequences to be considered a match
        :return: True if chain, v_gene and j_gene are the same and sequences are within given Levenshtein distance
        """
        return reference_sequence.locus == original_sequence.locus \
            and self.matches_gene(reference_sequence.v_call, original_sequence.v_call) \
            and self.matches_gene(reference_sequence.j_call, original_sequence.j_call) \
            and edit_distance(original_sequence.get_sequence(), reference_sequence.get_sequence()) <= max_distance

    def match_repertoire(self, repertoire: Repertoire, index: int, reference_sequences: list, max_distance: int,
                         summary_type: SequenceMatchingSummaryType, num_processes: int = 1) -> dict:

        matched = {"sequences": [], "repertoire": repertoire.identifier, "repertoire_index": index}
        arguments = [(seq, reference_sequences, max_distance) for seq in repertoire.sequences()]

        with Pool(num_processes) as pool:
            matched["sequences"] = pool.starmap(self.match_sequence, arguments)

        if summary_type == SequenceMatchingSummaryType.CLONAL_PERCENTAGE:
            total_count = np.sum([sequence.duplicate_count for sequence in repertoire.sequences()])
            matched["clonal_percentage"] = np.sum([sequence.duplicate_count for index, sequence in enumerate(repertoire.sequences()) if len(matched["sequences"][index]["matching_sequences"]) > 0]) / total_count
        else:
            matched["count"] = len([r for r in matched["sequences"] if len(r["matching_sequences"]) > 0])
            matched["percentage"] = matched["count"] / len(matched["sequences"])
        matched["metadata"] = repertoire.metadata
        matched["patient_id"] = repertoire.identifier
        matched["locus"] = list(set([sequence.locus for sequence in repertoire.sequences()]))

        return matched

    def match_sequence(self, sequence: ReceptorSequence, reference_sequences: list, max_distance: int) -> dict:
        matching_sequences = [seq.get_sequence() for seq in reference_sequences
                              if self.matches_sequence(sequence, seq, max_distance)]

        return {
            "matching_sequences": matching_sequences,
            "sequence": sequence.get_sequence(),
            "v_call": sequence.v_call,
            "j_call": sequence.j_call,
            "locus": sequence.locus
        }



from pathlib import Path
from typing import Tuple

import numpy as np
import plotly.graph_objects as go
from scipy.stats import beta

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_methods.classifiers.ProbabilisticBinaryClassifier import ProbabilisticBinaryClassifier
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.util.PathBuilder import PathBuilder


class SequenceAssociationLikelihood(MLReport):
    """
    Plots the beta distribution used as a prior for class assignment in ProbabilisticBinaryClassifier. The distribution plotted shows
    the probability that a sequence is associated with a given class for a label.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_sequence_assoc_report: SequenceAssociationLikelihood

    """

    DISTRIBUTION_PERCENTAGE_TO_SHOW = 0.999
    STEP = 400

    @classmethod
    def build_object(cls, **kwargs):
        return SequenceAssociationLikelihood(**kwargs)

    def check_prerequisites(self):
        if not isinstance(self.method, ProbabilisticBinaryClassifier):
            return False
        else:
            return True

    def __init__(self, train_dataset: Dataset = None, test_dataset: Dataset = None,
                 method: MLMethod = None, result_path: Path = None, name: str = None, hp_setting: HPSetting = None,
                 label=None, number_of_processes: int = 1):
        super().__init__(train_dataset=train_dataset, test_dataset=test_dataset, method=method, result_path=result_path,
                         name=name, hp_setting=hp_setting, label=label, number_of_processes=number_of_processes)
        self.result_name = None

    def _generate(self) -> ReportResult:

        PathBuilder.build(self.result_path)

        lower_limit, upper_limit = self.get_distribution_limits()
        self.result_name = "beta_distribution"

        report_output_fig = self._plot(upper_limit=upper_limit, lower_limit=lower_limit)
        output_figures = [] if report_output_fig is None else [report_output_fig]

        return ReportResult(name=self.name,
                            info="Beta distribution priors - probability that a sequence is disease-associated",
                            output_figures=output_figures)

    def _plot(self, upper_limit, lower_limit):

        beta_distribution_x = np.arange(start=lower_limit, stop=upper_limit, step=(upper_limit - lower_limit) / SequenceAssociationLikelihood.STEP)
        negative_pdf = beta.pdf(beta_distribution_x, self.method.alpha_0, self.method.beta_0)
        positive_pdf = beta.pdf(beta_distribution_x, self.method.alpha_1, self.method.beta_1)

        figure = go.Figure()
        figure.add_trace(go.Scatter(x=beta_distribution_x, y=negative_pdf, mode='lines', line=dict(color='#E69F00', width=2), name=f"{self.method.get_label_name()} {self.method.class_mapping[0]}"))
        figure.add_trace(go.Scatter(x=beta_distribution_x, y=positive_pdf, mode='lines', line=dict(color='#0072B2', width=2), name=f"{self.method.get_label_name()} {self.method.class_mapping[1]}"))

        figure.update_layout(template="plotly_white", xaxis_title=f"probability that receptor sequence is {self.method.get_label_name()}-associated",
                             yaxis_title="probability density function", xaxis={'tickformat': '.2e'}, yaxis={'tickformat': '.2e'})

        output_path = self.result_path / f"{self.result_name}.html"

        output_path = PlotlyUtil.write_image_to_file(figure, output_path)

        return ReportOutput(output_path)

    def get_distribution_limits(self) -> Tuple[float, float]:
        lower_limit_0, upper_limit_0 = beta.interval(SequenceAssociationLikelihood.DISTRIBUTION_PERCENTAGE_TO_SHOW,
                                                     self.method.alpha_0, self.method.beta_0)
        lower_limit_1, upper_limit_1 = beta.interval(SequenceAssociationLikelihood.DISTRIBUTION_PERCENTAGE_TO_SHOW,
                                                     self.method.alpha_1, self.method.beta_1)
        lower_limit = min(lower_limit_0, lower_limit_1)
        upper_limit = max(upper_limit_0, upper_limit_1)

        return lower_limit, upper_limit



import numpy as np

from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.SequenceParams import RegionType, Chain
from immuneML.data_model.SequenceSet import ReceptorSequence
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import SequenceDataset, ReceptorDataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.util.EncoderHelper import EncoderHelper
from immuneML.util.ParameterValidator import ParameterValidator


class SillyEncoder(DatasetEncoder):
    """
    This SillyEncoder class is a placeholder for a real encoder.
    It computes a set of random numbers as features for a given dataset.

    **Dataset type:**

    - SequenceDatasets

    - ReceptorDatasets

    - RepertoireDatasets


    **Specification arguments:**

    - random_seed (int): The random seed for generating random features.

    - embedding_len (int): The number of random features to generate per example.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_silly_encoder:
                    Silly: # name of the class (without 'Encoder' suffix)
                        random_seed: 1
                        embedding_len: 5
    """

    def __init__(self, random_seed: int, embedding_len: int, name: str = None):
        # The encoder name contains the user-defined name for the encoder. It may be used by some reports.
        super().__init__(name=name)

        # All user parameters are set here.
        # Default parameters must not be defined in the Encoder class, but in a default parameters file.
        self.random_seed = random_seed
        self.embedding_len = embedding_len

    @staticmethod
    def build_object(dataset=None, **params):
        # build_object is called early in the immuneML run, before the analysis takes place.
        # Its purpose is to fail early when a class is called incorrectly (checking parameters and dataset),
        # and provide user-friendly error messages.

        # ParameterValidator contains many utility functions for checking user parameters
        ParameterValidator.assert_type_and_value(params['random_seed'], int, SillyEncoder.__name__, 'random_seed', min_inclusive=1)
        ParameterValidator.assert_type_and_value(params['embedding_len'], int, SillyEncoder.__name__, 'embedding_len', min_inclusive=1, max_inclusive=100)

        # An error should be thrown if the dataset type is incompatible with the Encoder.
        # If different sub-classes are defined for each dataset type (e.g., OneHotRepertoireEncoder),
        # an instance of the dataset-specific class must be returned here.
        if isinstance(dataset, SequenceDataset) or isinstance(dataset, ReceptorDataset) or isinstance(dataset, RepertoireDataset):
            return SillyEncoder(**params)
        else:
            raise ValueError("SillyEncoder is only defined for dataset types SequenceDataset, ReceptorDataset or RepertoireDataset")

    def encode(self, dataset, params: EncoderParams) -> Dataset:
        np.random.seed(self.random_seed)

        # Generate the design matrix from the sequence dataset
        encoded_examples = self._get_encoded_examples(dataset)

        # EncoderHelper contains some utility functions, including this function for encoding the labels
        labels = EncoderHelper.encode_dataset_labels(dataset, params.label_config, params.encode_labels)

        # Each feature is represented by some meaningful name
        feature_names = [f"random_number_{i}" for i in range(self.embedding_len)]

        encoded_data = EncodedData(examples=encoded_examples,
                                   labels=labels,
                                   example_ids=dataset.get_example_ids(),
                                   feature_names=feature_names,
                                   encoding=SillyEncoder.__name__) # When using dataset-specific encoders,
                                                                   # make sure to use the general encoder name here
                                                                   # (e.g., OneHotEncoder.__name__, not OneHotSequenceEncoder.__name__)

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = encoded_data

        return encoded_dataset

    def _get_encoded_examples(self, dataset: Dataset) -> np.array:
        if isinstance(dataset, SequenceDataset):
            return self._get_encoded_sequences(dataset)
        elif isinstance(dataset, ReceptorDataset):
            return self._get_encoded_receptors(dataset)
        elif isinstance(dataset, RepertoireDataset):
            return self._get_encoded_repertoires(dataset)

    def _get_encoded_sequences(self, dataset: SequenceDataset) -> np.array:

        # for sequence dataset, the data can be access as a list of ReceptorSequence objects:
        for sequence in dataset.get_data():
            # Each sequence is a ReceptorSequence object.
            # Different properties of the sequence can be retrieved here, examples:
            identifier = sequence.sequence_id
            aa_seq = sequence.sequence_aa
            v_gene = sequence.v_call
            j_gene = sequence.j_call

        # alternatively, data can be accessed through the AIRRSequenceSet object
        # (a bionumpy dataclass, similar to pandas dataframes):

        all_cdr3_aa = dataset.data.cdr3_aa.tolist()  # a list of strings of all sequences in the dataset
        all_v_call = dataset.data.v_call.tolist()  # a list of strings of all v genes in the dataset

        # it can also be accessed as a bionumpy dataclass object:
        cdr3_aa = dataset.data.cdr3_aa  # useful for faster computations

        # in this example, we just return a random matrix of n_examples x n_features
        return np.random.rand(dataset.element_count, self.embedding_len)

    def _get_encoded_receptors(self, dataset: ReceptorDataset) -> np.array:

        # for receptor dataset, the data can be access as a list of Receptor objects:
        for receptor in dataset.get_data():
            # A Receptor contains two paired ReceptorSequence objects
            identifier = receptor.receptor_id
            chain1, chain2 = receptor.get_chains()
            sequence1 = receptor.get_chain(chain1)
            sequence2 = receptor.get_chain(chain2)

            # Properties of the specific ReceptorSequences can be retrieved, examples:
            aa_seq1 = sequence1.sequence_aa # gets the amino acid sequence by default (alternative: nucleotide)
            v_gene_seq1 = sequence1.v_call # gets the v and j genes
            j_gene_seq1 = sequence1.j_call

        # data can be also accessed through the AIRRSequenceSet object
        # (a bionumpy dataclass, similar to pandas dataframes)
        # this is useful when speed is important

        # all cdr3 aa sequences of the beta chain
        cdr3_aa_beta = dataset.data.cdr3_aa[dataset.data.locus == Chain.BETA.to_string()].tolist()
        # all cdr3 aa sequences of the alpha chain
        cdr3_aa_alpha = dataset.data.cdr3_aa[dataset.data.locus == Chain.ALPHA.to_string()].tolist()

        # here we just return a random matrix of n_examples x n_features
        return np.random.rand(dataset.element_count, self.embedding_len)

    def _get_encoded_repertoires(self, dataset: RepertoireDataset) -> np.array:
        encoded_repertoires = []

        for repertoire in dataset.get_data():
            # Each repertoire is a Repertoire object.
            # Receptor sequences from the repertoire objects can be retrieved as objects for the given region type:
            for sequence in repertoire.sequences(region_type=RegionType.IMGT_CDR3):
                assert isinstance(sequence, ReceptorSequence) # each sequence is an object of ReceptorSequence class
                seq_id = sequence.sequence_id
                aa_seq = sequence.sequence_aa
                v_call = sequence.v_call

            # alternatively, the sequences can be retrieved as AIRRSequenceSet objects (bionumpy dataclasses,
            # similar to pandas dataframes):
            all_aa_sequences = repertoire.data.cdr3_aa.tolist()  # a list of strings of all sequences in the repertoire
            all_aa_sequences = repertoire.data.cdr3_aa  # a bionumpy dataclass object containing all sequences in the repertoire

            # other properties can also be accessed in the same way:
            v_calls = repertoire.data.v_call.tolist()  # a list of strings of all v genes in the repertoire

            # In this encoding, repertoire information is ignored, random features are generated
            random_encoding = np.random.rand(self.embedding_len)
            encoded_repertoires.append(random_encoding)

        return np.array(encoded_repertoires)


from pathlib import Path

from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.presentation.html.HTMLBuilder import HTMLBuilder
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class OutputParser:

    @staticmethod
    def parse(specs: dict, symbol_table: SymbolTable) -> dict:
        if "output" in specs:
            ParameterValidator.assert_keys(specs["output"], ["format"], "OutputParser", "output")
            ParameterValidator.assert_in_valid_list(specs["output"]["format"], ["HTML"], "OutputParser", "format")
        else:
            specs["output"] = {"format": "HTML"}
        symbol_table.add("output", SymbolType.OUTPUT, specs["output"])

        return specs["output"]

    @staticmethod
    def generate_docs(path: Path):
        output_path = PathBuilder.build(path / "output")
        output_path = output_path / "outputs.rst"
        with output_path.open("w") as file:
            file.writelines("\n".join([el.replace('    ', '', 1) if el.startswith('    ') else el for el in
                                       HTMLBuilder.__doc__.split("\n")]))



import pickle

import numpy as np

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.util.PathBuilder import PathBuilder


class AbundanceEncoderHelper:
    INVALID_P_VALUE = 2.0

    @staticmethod
    def check_is_positive_class(dataset, matrix_repertoire_ids, label_config: LabelConfiguration):
        label = label_config.get_label_objects()[0]

        is_positive_class = np.array(
            [dataset.get_repertoire(repertoire_identifier=repertoire_id).metadata[label.name] for repertoire_id in
             matrix_repertoire_ids]) == label.positive_class

        return is_positive_class

    @staticmethod
    def get_relevant_sequence_indices(sequence_presence_iterator, is_positive_class, p_value_threshold,
                                      relevant_indices_path, params,
                                      cache_params=None):

        import fisher

        relevant_indices_path = relevant_indices_path if relevant_indices_path is not None else params.result_path / 'relevant_sequence_indices' \
                                                                                                                     '.pickle '
        file_paths = {"relevant_indices_path": relevant_indices_path}

        if params.learn_model:
            contingency_table = CacheHandler.memo_by_params(
                ('cache_params', cache_params, ('type', 'contingency_table')),
                lambda: AbundanceEncoderHelper._get_contingency_table(sequence_presence_iterator,
                                                                      is_positive_class))
            p_values = CacheHandler.memo_by_params((('cache_params', cache_params), ("type", "fisher_p_values")),
                                                   lambda: AbundanceEncoderHelper._find_sequence_p_values_with_fisher(
                                                       contingency_table, fisher))
            relevant_sequence_indices = p_values < p_value_threshold

            file_paths["contingency_table_path"] = AbundanceEncoderHelper._write_contingency_table(contingency_table,
                                                                                                   params.result_path)
            file_paths["p_values_path"] = AbundanceEncoderHelper._write_p_values(p_values, params.result_path)

            with relevant_indices_path.open("wb") as file:
                pickle.dump(relevant_sequence_indices, file)
        else:
            with relevant_indices_path.open("rb") as file:
                relevant_sequence_indices = pickle.load(file)

        return relevant_sequence_indices, file_paths

    @staticmethod
    def _get_contingency_table(sequence_presence_iterator, is_positive_class):
        contingency_table = np.zeros(shape=(len(sequence_presence_iterator), 4), dtype=int)

        for i, sequence_vector in enumerate(sequence_presence_iterator):
            contingency_table[i, 0] = np.sum(sequence_vector[np.logical_and(sequence_vector, is_positive_class)])
            contingency_table[i, 1] = np.sum(
                sequence_vector[np.logical_and(sequence_vector, np.logical_not(is_positive_class))])
            contingency_table[i, 2] = np.sum(np.logical_and(is_positive_class, sequence_vector == 0))
            contingency_table[i, 3] = np.sum(np.logical_and(np.logical_not(is_positive_class), sequence_vector == 0))

        return contingency_table

    @staticmethod
    def _find_sequence_p_values_with_fisher(contingency_table, fisher):
        fisher_func = AbundanceEncoderHelper.make_fisher_function(fisher)
        return np.apply_along_axis(fisher_func, 1, contingency_table)

    @staticmethod
    def make_fisher_function(fisher):
        def fisher_func(row_in):
            if row_in[0] + row_in[1] > 1:
                return fisher.pvalue(row_in[0], row_in[1], row_in[2], row_in[3]).right_tail
            else:
                return AbundanceEncoderHelper.INVALID_P_VALUE

        return fisher_func

    @staticmethod
    def _write_contingency_table(contingency_table, result_path):
        contingency_table_path = PathBuilder.build(result_path) / 'contingency_table.csv'

        np.savetxt(contingency_table_path, contingency_table, fmt="%s", delimiter=",",
                   header="positive_present,negative_present,positive_absent,negative_absent", comments='')

        return contingency_table_path

    @staticmethod
    def _write_p_values(p_values, result_path):
        p_values_path = result_path / 'p_values.csv'

        np.savetxt(p_values_path, p_values, header="p_values", comments='')

        return p_values_path

    @staticmethod
    def build_abundance_matrix(sequence_presence_matrix, matrix_repertoire_ids, dataset_repertoire_ids,
                               sequence_p_values_indices):
        abundance_matrix = np.zeros((len(dataset_repertoire_ids), 2))

        for idx_in_dataset, dataset_repertoire_id in enumerate(dataset_repertoire_ids):
            relevant_row = np.where(matrix_repertoire_ids == dataset_repertoire_id)
            repertoire_vector = sequence_presence_matrix.T[relevant_row]
            relevant_sequence_abundance = np.sum(
                repertoire_vector[np.logical_and(sequence_p_values_indices, repertoire_vector)])
            total_sequence_abundance = np.sum(repertoire_vector)
            abundance_matrix[idx_in_dataset] = [relevant_sequence_abundance, total_sequence_abundance]

        return abundance_matrix

    @staticmethod
    def get_matching_func_for_repertoire(comparison_attributes: list):
        def matching_func(repertoire):
            tmp_df = repertoire.data.topandas()[comparison_attributes]
            tmp_df = tmp_df[tmp_df.astype(bool).any(axis=1)]
            unique_tuples = list(tmp_df.drop_duplicates().itertuples(index=False, name=None))

            return unique_tuples

        return matching_func



# quality: peripheral
import datetime
import re
from pathlib import Path

import yaml
from yaml import MarkedYAMLError

from immuneML.dsl.InstructionParser import InstructionParser
from immuneML.dsl.OutputParser import OutputParser
from immuneML.dsl.definition_parsers.DefinitionParser import DefinitionParser
from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder


class ImmuneMLParser:
    """
    Simple DSL parser from python dictionary or equivalent YAML for configuring repertoire / receptor_sequence
    classification in the (simulated) settings

    DSL example with hyper-parameter optimization:

    .. highlight:: yaml
    .. code-block:: yaml

        definitions:
            datasets:
                d1:
                    format: MiXCR
                    params:
                        result_path: loaded_dataset/
                        region_type: IMGT_CDR3
                        path: path_to_files/
                        metadata_file: metadata.csv
            encodings:
                e1:
                    KmerFrequency
                        k: 3
                e2:
                    Word2Vec:
                        vector_size: 16
                        context: sequence
            ml_methods:
                log_reg1:
                    LogisticRegression:
                        C: 0.001
            reports:
                r1:
                    SequenceLengthDistribution
            preprocessing_sequences:
                seq1:
                    - filter_chain_B:
                        ChainRepertoireFilter:
                            keep_chain: A
                    - filter_clonotype:
                        ClonesPerRepertoireFilter:
                            lower_limit: 1000
                seq2:
                    - filter_clonotype:
                        ClonesPerRepertoireFilter:
                            lower_limit: 500
                    - filter_chain_A:
                        ChainRepertoireFilter:
                            keep_chain: B
        instructions:
            inst1:
                type: TrainMLModel
                settings:
                    -   preprocessing: seq1
                        encoding: e1
                        ml_method: log_reg1
                    -   preprocessing: seq2
                        encoding: e2
                        ml_method: log_reg1
                assessment:
                    split_strategy: random
                    split_count: 1
                    training_percentage: 70
                    reports:
                        data: []
                        data_splits: []
                        encoding: []
                        models: []
                selection:
                    split_strategy: k-fold
                    split_count: 5
                    reports:
                        data: []
                        data_splits: [r1]
                        encoding: []
                        models: []
                labels:
                    - CD
                dataset: d1
                strategy: GridSearch
                metrics: [accuracy, f1_micro]
                optimization_metric: balanced_accuracy
                reports: []
        output: # this section can also be omitted, in that case output will be automatically HTML
            format: HTML # or None

    """

    @staticmethod
    def parse_yaml_file(file_path: Path, result_path: Path = None, parse_func=None):
        try:
            with file_path.open("r") as file:
                workflow_specification = yaml.safe_load(file)
                ImmuneMLParser.check_keys(workflow_specification)
        except yaml.YAMLError as exc:
            problem_description = "\n--------------------------------------------------------------------------------\n" \
                                  "There was a YAML formatting error in the supplied specification file. Please validate specification " \
                                  "(you can use https://jsonformatter.org/yaml-validator) and try again."
            raise MarkedYAMLError(context=str(exc), problem=problem_description, problem_mark=f"The error was {exc.problem_mark}.")

        try:
            if parse_func is None:
                symbol_table, path = ImmuneMLParser.parse(workflow_specification, file_path, result_path)
            else:
                symbol_table, path = parse_func(workflow_specification, file_path, result_path)
        except KeyError as key_error:
            raise Exception(f"ImmuneMLParser: an error occurred during parsing the YAML specification. Missing key was '{key_error.args[0]}'. "
                            f"For more details, refer to the log above and check the documentation.") from key_error
        return symbol_table, path

    @staticmethod
    def check_keys(specs: dict):
        for key in specs.keys():
            key_to_check = str(key)
            assert re.match(r'^[A-Za-z0-9_]+$', key_to_check), \
                f"ImmuneMLParser: the keys in the specification can contain only letters, numbers and underscore. Error with key: {key}"
            if isinstance(specs[key], dict) and key not in ["column_mapping", "metadata_column_mapping"]:
                ImmuneMLParser.check_keys(specs[key])

    @staticmethod
    def parse(workflow_specification: dict, file_path, result_path):

        symbol_table = SymbolTable()

        def_parser_output, specs_defs = DefinitionParser.parse(workflow_specification, symbol_table, result_path)
        symbol_table, specs_instructions = InstructionParser.parse(def_parser_output, result_path)
        app_output = OutputParser.parse(workflow_specification, symbol_table)

        path = ImmuneMLParser._output_specs(file_path=file_path, result_path=result_path, definitions=specs_defs,
                                            instructions=specs_instructions, output=app_output)

        return symbol_table, path

    @staticmethod
    def _get_full_specs_filepath(file_path, result_path) -> Path:
        file_name = f"full_{file_path.stem}.yaml"
        if result_path is None:
            folder = file_path.absolute().parent
            return folder / file_name
        else:
            return result_path / file_name

    @staticmethod
    def _output_specs(file_path=None, result_path=None, definitions: dict = None, instructions: dict = None, output: dict = None) -> Path:
        filepath = ImmuneMLParser._get_full_specs_filepath(file_path, result_path)

        result = {"definitions": definitions, "instructions": instructions, "output": output}
        result = ImmuneMLParser._paths_to_strings_recursive(result)

        PathBuilder.build(filepath.parent)
        with filepath.open("w") as file:
            yaml.dump(result, file)

        print_log(f"Full specification is available at {filepath}.\n", include_datetime=True)
        return filepath

    @staticmethod
    def _paths_to_strings_recursive(specs):
        if isinstance(specs, Path):
            return specs.as_posix()
        elif isinstance(specs, dict):
            return {key: ImmuneMLParser._paths_to_strings_recursive(value) for key, value in specs.items()}
        elif isinstance(specs, list):
            return [ImmuneMLParser._paths_to_strings_recursive(item) for item in specs]
        else:
            return specs




from inspect import signature
from pathlib import Path

from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.dsl.definition_parsers.DefinitionParserOutput import DefinitionParserOutput
from immuneML.dsl.definition_parsers.EncodingParser import EncodingParser
from immuneML.dsl.definition_parsers.ExampleWeightingParser import ExampleWeightingParser
from immuneML.dsl.definition_parsers.MLParser import MLParser
from immuneML.dsl.definition_parsers.MotifParser import MotifParser
from immuneML.dsl.definition_parsers.PreprocessingParser import PreprocessingParser
from immuneML.dsl.definition_parsers.ReportParser import ReportParser
from immuneML.dsl.definition_parsers.SignalParser import SignalParser
from immuneML.dsl.definition_parsers.SimulationParser import SimulationParser
from immuneML.dsl.import_parsers.ImportParser import ImportParser
from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.protein_embedding.ProteinEmbeddingEncoder import ProteinEmbeddingEncoder
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_methods.clustering.ClusteringMethod import ClusteringMethod
from immuneML.ml_methods.dim_reduction.DimRedMethod import DimRedMethod
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.preprocessing.Preprocessor import Preprocessor
from immuneML.reports.clustering_method_reports.ClusteringMethodReport import ClusteringMethodReport
from immuneML.reports.clustering_reports.ClusteringReport import ClusteringReport
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.reports.gen_model_reports.GenModelReport import GenModelReport
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.reports.multi_dataset_reports.MultiDatasetReport import MultiDatasetReport
from immuneML.reports.train_ml_model_reports.TrainMLModelReport import TrainMLModelReport
from immuneML.simulation.SimConfig import SimConfig
from immuneML.simulation.SimConfigItem import SimConfigItem
from immuneML.simulation.implants.LigoPWM import LigoPWM
from immuneML.simulation.implants.Motif import Motif
from immuneML.simulation.implants.SeedMotif import SeedMotif
from immuneML.simulation.implants.Signal import Signal
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.ReflectionHandler import ReflectionHandler
from scripts.DocumentatonFormat import DocumentationFormat
from scripts.specification_util import write_class_docs, make_docs


class DefinitionParser:

    @staticmethod
    def parse(workflow_specification: dict, symbol_table: SymbolTable, result_path: Path):

        specs = workflow_specification["definitions"]

        specs_defs = {}

        for parser in [MotifParser, SignalParser, SimulationParser, PreprocessingParser, EncodingParser, ExampleWeightingParser,
                       MLParser, ReportParser, ImportParser]:
            symbol_table, new_specs = DefinitionParser._call_if_exists(parser.keyword, parser.parse, specs,
                                                                       symbol_table, result_path)
            specs_defs[parser.keyword] = new_specs

        return DefinitionParserOutput(symbol_table=symbol_table, specification=workflow_specification), specs_defs

    @staticmethod
    def _call_if_exists(key: str, method, specs: dict, symbol_table: SymbolTable, path=None):
        if key in specs:
            if "path" in signature(method).parameters:
                return method(specs[key], symbol_table, path)
            else:
                return method(specs[key], symbol_table)
        else:
            return symbol_table, {}

    @staticmethod
    def generate_docs(path: Path):
        def_path = PathBuilder.build(path / "definitions")
        DefinitionParser.make_dataset_docs(def_path)
        DefinitionParser.make_simulation_docs(def_path)
        DefinitionParser.make_encodings_docs(def_path)
        DefinitionParser.make_reports_docs(def_path)
        DefinitionParser.make_ml_methods_docs(def_path)
        DefinitionParser.make_preprocessing_docs(def_path)

    @staticmethod
    def make_simulation_docs(path: Path):
        classes_to_document = [DocumentationFormat(Motif, "Motifs", DocumentationFormat.LEVELS[1]),
                               DocumentationFormat(SeedMotif, SeedMotif.__name__, DocumentationFormat.LEVELS[2]),
                               DocumentationFormat(LigoPWM, "PWM", DocumentationFormat.LEVELS[2]),
                               DocumentationFormat(Signal, "Signals", DocumentationFormat.LEVELS[1]),
                               DocumentationFormat(SimConfig, "Simulation config", DocumentationFormat.LEVELS[1]),
                               DocumentationFormat(SimConfigItem, "Simulation config item", DocumentationFormat.LEVELS[2])]

        file_path = path / "specs_simulation.rst"
        with file_path.open("w") as file:
            for doc_format in classes_to_document:
                write_class_docs(doc_format, file)

    @staticmethod
    def make_dataset_docs(path: Path):
        import_classes = ReflectionHandler.all_nonabstract_subclasses(DataImport, "Import", "dataset_import/")
        make_docs(path, import_classes, "specs_datasets.rst", "Import")

    @staticmethod
    def make_encodings_docs(path: Path):
        enc_classes = ReflectionHandler.all_direct_subclasses(DatasetEncoder, "Encoder", "encodings/")
        enc_classes = [cl for cl in enc_classes if cl.__name__ != "ProteinEmbeddingEncoder"]
        enc_classes += ReflectionHandler.all_direct_subclasses(ProteinEmbeddingEncoder, "Encoder", "encodings/")

        make_docs(path, enc_classes, "specs_encodings.rst", "Encoder")

    @staticmethod
    def make_reports_docs(path: Path):
        filename = "specs_reports.rst"
        file_path = path / filename
        mode = "w"

        report_types = [DataReport, EncodingReport, MLReport, GenModelReport, TrainMLModelReport, ClusteringReport,
                        MultiDatasetReport]

        for report_type_class in report_types:
            with file_path.open(mode) as file:
                doc_format = DocumentationFormat(cls=report_type_class,
                                                 cls_name=f"**{report_type_class.DOCS_TITLE}**",
                                                 level_heading=DocumentationFormat.LEVELS[1])
                write_class_docs(doc_format, file)
                mode = "a"

            subdir = DefaultParamsLoader.convert_to_snake_case(report_type_class.__name__) + "s"

            classes = ReflectionHandler.all_nonabstract_subclasses(report_type_class, "", f"reports/{subdir}/")
            make_docs(path, classes, filename, "", "a", format_level=DocumentationFormat.LEVELS[2])

    @staticmethod
    def make_ml_methods_docs(path: Path):
        filename = 'specs_ml_methods.rst'
        file_path = path / filename
        mode = "w"

        method_mapping = [{'method_type': MLMethod, 'subdir': 'classifiers', 'title': MLMethod.DOCS_TITLE},
                          {'method_type': ClusteringMethod, 'subdir': 'clustering', 'title': ClusteringMethod.DOCS_TITLE},
                          {'method_type': GenerativeModel, 'subdir': 'generative_models', 'title': GenerativeModel.DOCS_TITLE},
                          {'method_type': DimRedMethod, 'subdir': 'dim_reduction', 'title': DimRedMethod.DOCS_TITLE}]

        for method in method_mapping:
            with file_path.open(mode) as file:
                doc_format = DocumentationFormat(cls=method['method_type'],
                                                 cls_name=f"**{method['title']}**",
                                                 level_heading=DocumentationFormat.LEVELS[1])
                write_class_docs(doc_format, file)
                mode = "a"

            classes = ReflectionHandler.all_nonabstract_subclasses(method['method_type'], "",
                                                                   f"ml_methods/{method['subdir']}/")
            make_docs(path, classes, filename, "", "a", format_level=DocumentationFormat.LEVELS[2])

    @staticmethod
    def make_preprocessing_docs(path: Path):
        classes = ReflectionHandler.all_nonabstract_subclasses(Preprocessor, "", "preprocessing/")
        make_docs(path, classes, "specs_preprocessings.rst", "")



from collections import Counter

from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.kmer_frequency.KmerFrequencyEncoder import KmerFrequencyEncoder


class KmerFreqSequenceEncoder(KmerFrequencyEncoder):

    def _encode_locus(self, dataset):
        return len(set(dataset.data.locus.tolist())) > 1

    def _encode_new_dataset(self, dataset, params: EncoderParams):

        encoded_data = self._encode_data(dataset, params)

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = encoded_data

        return encoded_dataset

    def _encode_examples(self, dataset: SequenceDataset, params: EncoderParams):

        encoded_sequences = []
        sequence_ids = []
        label_config = params.label_config
        labels = {label: [] for label in label_config.get_labels_by_name()} if params.encode_labels else None

        encode_locus = self._encode_locus(dataset)

        sequence_encoder = self._prepare_sequence_encoder()
        params.region_type = self.region_type
        for sequence in dataset.get_data(region_type=self.region_type):
            counts = self._encode_sequence(sequence, params, sequence_encoder, Counter(), encode_locus)
            encoded_sequences.append(counts)
            sequence_ids.append(sequence.sequence_id)

            if params.encode_labels:
                for label_name in label_config.get_labels_by_name():
                    label = sequence.metadata[label_name]
                    labels[label_name].append(label)

        return encoded_sequences, sequence_ids, labels



import os
from pathlib import Path

from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.ml_methods.util.Util import Util as MLUtil
from immuneML.presentation.TemplateParser import TemplateParser
from immuneML.presentation.html.Util import Util


class FailedGalaxyHTMLBuilder:
    """
    Constructs the HTML file to display for failed Galaxy runs
    """

    CSS_PATH = EnvironmentSettings.html_templates_path / "css/custom.css"

    @staticmethod
    def build(result_path, traceback_str) -> Path:
        logfile = result_path / "log.txt"

        html_map = {
            "css_style": Util.get_css_content(FailedGalaxyHTMLBuilder.CSS_PATH),
            "logfile_content": open(logfile, 'r').read() if  os.path.isfile(logfile) else "Log file not found",
            "exception": traceback_str,
            'immuneML_version': MLUtil.get_immuneML_version()}

        result_file = result_path / f"index.html"

        TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "FailedGalaxy.html",
                             template_map=html_map, result_path=result_file)

        return result_file



from immuneML.analysis.criteria_matches.BooleanType import BooleanType
from immuneML.analysis.criteria_matches.OperationType import OperationType


class CriteriaTypeInstantiator:

    @staticmethod
    def instantiate(criteria):
        if criteria["type"].upper() in OperationType._member_names_:
            return {**criteria, "type": OperationType[criteria["type"].upper()]}
        elif criteria["type"].upper() in BooleanType._member_names_:
            operands = []
            for operand in criteria["operands"]:
                operands.append(CriteriaTypeInstantiator.instantiate(operand))
            return {**criteria, "operands": operands, "type": BooleanType[criteria["type"].upper()]}



import inspect

from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_methods.clustering.ClusteringMethod import ClusteringMethod
from immuneML.ml_methods.dim_reduction.DimRedMethod import DimRedMethod
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.util.Logger import log
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReflectionHandler import ReflectionHandler


class MLParser:
    keyword = "ml_methods"

    @staticmethod
    def parse(specification: dict, symbol_table: SymbolTable):

        for ml_method_id in specification.keys():
            ml_method, config = MLParser._parse_ml_method(ml_method_id, specification[ml_method_id])
            specification[ml_method_id] = config
            symbol_table.add(ml_method_id, SymbolType.ML_METHOD, ml_method, config)

        return symbol_table, specification

    @staticmethod
    def _parse_ml_method(ml_method_id: str, ml_specification) -> tuple:

        valid_class_values = {'classifiers': ReflectionHandler.all_nonabstract_subclass_basic_names(MLMethod, "",
                                                                                                    "ml_methods/classifiers/"),
                              'gen_models': ReflectionHandler.all_nonabstract_subclass_basic_names(GenerativeModel, "",
                                                                                                   "ml_methods/generative_models/"),
                              'clustering': ReflectionHandler.all_nonabstract_subclass_basic_names(ClusteringMethod, "",
                                                                                                   "ml_methods/clustering/"),
                              'dim_reduction': ReflectionHandler.all_nonabstract_subclass_basic_names(DimRedMethod, "",
                                                                                                      "ml_methods/dim_reduction/")}

        if type(ml_specification) is str:
            ml_specification = {ml_specification: {}}

        return MLParser.parse_by_type(ml_method_id, ml_specification, valid_class_values)

    @staticmethod
    def parse_by_type(ml_method_id, ml_specification, valid_class_values):
        ml_class_name = [key for key in ml_specification.keys() if
                         key not in ['model_selection_cv', 'model_selection_n_folds']]
        assert len(ml_class_name) == 1, f"MLParser: method under {ml_method_id} is missing class name."
        ml_class_name = ml_class_name[0]

        if ml_class_name in valid_class_values['classifiers']:
            return MLParser.parse_classifiers(ml_method_id, ml_specification, ml_class_name)
        elif any(ml_class_name in class_list for class_list in [valid_class_values['gen_models'],
                                                                valid_class_values['clustering'],
                                                                valid_class_values['dim_reduction']]):
            return MLParser.parse_any_model(ml_method_id, ml_specification, ml_class_name)

        raise AssertionError(f"MLParser: {ml_class_name} is not a valid class name for any type of ML methods defined "
                             f"in immuneML. Valid values per type are: {valid_class_values}.")

    @staticmethod
    def parse_any_model(ml_method_id, ml_specification, ml_class_name):
        ml_method_class = ReflectionHandler.get_class_by_name(ml_class_name, subdirectory='ml_methods/')

        ml_specification[ml_class_name] = {
            **DefaultParamsLoader.load("ml_methods/", ml_class_name, log_if_missing=False),
            **ml_specification[ml_class_name]}

        method, params = MLParser.create_method_instance(ml_specification, ml_method_class, ml_method_id)
        ml_specification[ml_class_name] = params
        method.name = ml_method_id

        return method, ml_specification

    @staticmethod
    def parse_classifiers(ml_method_id, ml_specification, ml_class_name):
        ml_specification = {**DefaultParamsLoader.load("ml_methods/", "MLMethod"), **ml_specification}
        ml_specification_keys = list(ml_specification.keys())

        non_default_keys = [key for key in ml_specification.keys() if
                            key not in ["model_selection_cv", "model_selection_n_folds"]]

        assert len(
            ml_specification_keys) == 3, f"MLParser: ML method {ml_method_id} was not correctly specified. Expected at least 1 key " \
                                         f"(ML method name), got {len(ml_specification_keys) - 2} instead: " \
                                         f"{str([key for key in non_default_keys])[1:-1]}."

        return MLParser.parse_any_model(ml_method_id, ml_specification, ml_class_name)

    @staticmethod
    def create_method_instance(ml_specification: dict, ml_method_class, key: str) -> tuple:

        ml_params = {}

        if ml_specification[ml_method_class.__name__] is None or len(
                ml_specification[ml_method_class.__name__].keys()) == 0:
            ml_method = ml_method_class()
        else:
            ml_params = ml_specification[ml_method_class.__name__]
            init_method_keys = inspect.signature(ml_method_class.__init__).parameters.keys()
            if any([isinstance(ml_params[key], list) for key in
                    ml_params.keys()]) and "parameter_grid" in init_method_keys:

                ParameterValidator.assert_type_and_value(ml_specification['model_selection_cv'], bool,
                                                         MLParser.__name__, f'{key}: model_selection_cv')
                assert ml_specification[
                           'model_selection_cv'] is True, f"MLParser: when running ML method {key} with a list of inputs, model_selection_cv must be True! " \
                                                          f"Set the parameters for {key} to single values (not lists) or set model_selection_cv to True and model_selection_n_folds to >= 2"

                ParameterValidator.assert_type_and_value(ml_specification['model_selection_n_folds'], int,
                                                         MLParser.__name__, f'{key}: model_selection_n_folds', 2)

                ml_method = ml_method_class(
                    parameter_grid={key: [ml_params[key]] if not isinstance(ml_params[key], list) else ml_params[key]
                                    for key in ml_params.keys()})

            elif len(init_method_keys) == 3 and all(
                    arg in init_method_keys for arg in ["parameters", "parameter_grid"]):
                ml_method = ml_method_class(parameters=ml_params)
            else:
                ml_method = ml_method_class(**ml_params)

        if hasattr(ml_method, 'name'):
            ml_method.name = key

        return ml_method, ml_params



import abc
import logging
from pathlib import Path

from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.util.Logger import print_log


class Report(metaclass=abc.ABCMeta):
    """
    This class defines what report classes should look like: they all have to inherit this class and implement the abstract methods: build_object()
    from parameters and generate() the report once all properties are set (in immuneML this will be taken care of by the instructions). If there are
    any prerequisites needed to run the report (e.g., check if all parameter values are properly set), the check_prerequisites function should be
    overwritten to reflect that and determine if everything is set before generate() is run. See specific functions for more details.

    """

    def __init__(self, name: str = None, result_path: Path = None, number_of_processes: int = 1):
        self.name = name
        self.result_path = result_path
        self.number_of_processes = number_of_processes

    @classmethod
    @abc.abstractmethod
    def build_object(cls, **kwargs):
        """
        Creates the object of the subclass of the Report class from the parameters so that it can be used in the analysis. Depending on the type of
        the report, the parameters provided here will be provided in parsing time, while the other necessary parameters (e.g., subset of the data from
        which the report should be created) will be provided at runtime. For more details, see specific direct subclasses of this class, describing
        different types of reports.

        Args:

            **kwargs: keyword arguments that will be provided by users in the specification (if immuneML is used as a command line tool) or in the
             dictionary when calling the method from the code, and which should be used to create the report object

        Returns:

            the object of the appropriate report class

        """
        pass

    @abc.abstractmethod
    def _generate(self) -> ReportResult:
        """
        The function that needs to be implemented by the Report subclasses which actually creates the report (figures, tables, text files), depending
        on the specific aim of the report. After checking all prerequisites (e.g., if all parameters were set properly), generate_report() will call
        this function and return its result.

        Returns:

            ReportResult object which encapsulates all outputs (figure, table, and text files) so that they can be conveniently linked to in the
            final output of instructions

        """
        pass

    def check_prerequisites(self) -> bool:
        """
        Checks prerequisites for the generation of the report of specific class (e.g., if the class of the MLMethod instance is the one required by
        the report, if the data has been encoded to make a report of encoded dataset). In the instructions in immuneML, this function is used to
        determine whether to call generate_report() in the specific situation. Each report subclass has its own set of prerequisites. If the report
        cannot be run, the information on this will be logged and the report skipped in the specific situation. No error will be raised. See
        subclasses of the class :py:obj:`~immuneML.workflows.instructions.Instruction.Instruction` for more information on how the reports are executed.

        Returns:
             boolean value True if the prerequisites are o.k., and False otherwise.
        """
        return True

    def set_context(self, context: dict):
        """
        Context is a dictionary with information that is accessible from the level of instruction and can be used to precompute certain values that
        can be later reused to speed up the generation of the subsequent reports of the same time. For instance, if one should compute the distance
        between all repertoires based on the sequence content, it is possible to store the full dataset in the context, compute the distances on the
        full dataset and then only extract the distances need for the current dataset in the later calls (e.g., when training dataset is passed as
        input). Only some reports will need this functionality.

        Warning: It is very important to be careful when using the context to avoid leaking the information between training and test datasets.

        Args:

            context (dict): a dictionary where the values are variables that are typically only available on the top-level of the instruction, and
                which are used to precompute results in order to speed up subsequent generation of the same report on subsets of those values.

        Returns:

            self - so that it can be chained with the other function calls

        """
        return self

    def generate_report(self) -> ReportResult:
        """
        Generates a report of the given class if the prerequisites are satisfied. It handles all exceptions so that if there is an error while
        generating a report, the execution of the rest of the code (e.g., more time-expensive parts, like instructions) is not influenced.

        Returns:

            ReportResult object which encapsulates all outputs (figure, table, and text files) so that they can be conveniently linked to in the
            final output of instructions

        # """
        try:
            if self.check_prerequisites():
                return self._generate()
            else:
                print_log(
                    f"Report {self.name} failed while checking the prerequisites and was not generated, see the "
                    f"log file for more information.",
                    include_datetime=True)
                return ReportResult(name=f"{self.name} (failed)",
                                    info="This report failed while checking the prerequisites. This usually means "
                                         "the wrong dataset type was used, that wrong input parameters were specified "
                                         "or that the specific report is not applicable in this case. "
                                         "See the log file for more information.")
        except Exception as e:
            logging.exception(f"An exception occurred while generating report {self.name}. See the details below:")
            print_log(
                f"Report {self.name} encountered an error and could not be generated (error: {e}). "
                f"See the log file for more info.",
                include_datetime=True)
            return ReportResult(name=f"{self.name} (failed)",
                                info="This report failed during execution, see the log file for more information.")

    def _safe_plot(self, output_written=True, plot_callable="_plot", **kwargs):
        """
        A wrapper around the function _plot() which catches any error that may be thrown by this function (e.g. errors in R),
        and shows an informative warning message instead. This is to prevent immuneML from crashing when the analysis has been
        completed but only a figure could not be plotted.

        Args:

            output_written: indicates whether the output was written to a file, this changes the error message.

            kwargs: passed to _plot()

        Returns:

            the results of _plot(), typically a ReportOutput object

        """
        warning_mssg = f"{self.__class__.__name__}: an error occurred when attempting to plot the data. \n"
        if output_written:
            warning_mssg += "\nThe data has been written to a file, but no plot has been created."
        else:
            warning_mssg += "\nNo plot has been created."
        try:
            plot = getattr(self, plot_callable, None)
            if callable(plot):
                return plot(**kwargs)
        except Exception as e:
            logging.exception(
                f"An exception occurred while plotting the data in report {self.name}. See the details below:")
            print_log(warning_mssg, include_datetime=True)

    def _write_output_table(self, table, file_path, name=None):
        sep = "," if file_path.suffix == ".csv" else "\t"
        table.to_csv(file_path, index=False, sep=sep)

        return ReportOutput(path=file_path, name=name)



from dataclasses import dataclass, field
from typing import List, Dict, Union

from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.simulation.implants.Signal import Signal, SignalPair


@dataclass
class SimConfigItem:
    """
    When performing a simulation, one or more simulation config items can be specified.
    Config items define groups of repertoires or receptors that have the same simulation parameters,
    such as signals, generative model, clonal frequencies, noise parameters.


    **Specification arguments:**

    - signals (dict): signals for the simulation item and the proportion of sequences in the repertoire that will have the given signal. For receptor-level simulation, the proportion will always be 1.

    - is_noise (bool): indicates whether the implanting should be regarded as noise; if it is True, the signals will be implanted as specified, but the repertoire/receptor in question will have negative class.

    - generative_model: parameters of the generative model, including its type, path to the model; currently supported models are OLGA and ExperimentalImport

    - seed (int): starting random seed for the generative model (it should differ across simulation items, or it can be set to null when not used)

    - false_positives_prob_in_receptors (float): when performing repertoire level simulation, what percentage of sequences should be false positives

    - false_negative_prob_in_receptors (float): when performing repertoire level simulation, what percentage of sequences should be false negatives

    - immune_events (dict): a set of key-value pairs that will be added to the metadata (same values for all data generated in one simulation sim_item) and can be later used as labels

    - default_clonal_frequency (dict): clonal frequency in Ligo is simulated through `scipy's zeta distribution function for generating random numbers <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.zipf.html>`_, with parameters provided under default_clonal_frequency parameter. These parameters will be used to assign count values to sequences that do not contain any signals if they are required by the simulation. If clonal frequency shouldn't be used, this parameter can be None

    .. indent with spaces
    .. code-block:: yaml

        clonal_frequency:
            a: 2 # shape parameter of the distribution
            loc: 0 # 0 by default but can be used to shift the distribution

    - sequence_len_limits (dict): allows for filtering the generated sequences by length, needs to have parameters min and max specified; if not used, min/max should be -1

    .. indent with spaces
    .. code-block:: yaml

        sequence_len_limits:
            min: 4 # keep sequences of length 4 and longer
            max: -1 # no limit on the max length of the sequences

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            simulations: # definitions of simulations should be under key simulations in the definitions part of the specification
                # one simulation with multiple implanting objects, a part of definition section
                my_simulation:
                    sim_item1:
                        number_of_examples: 10
                        seed: null # don't use seed
                        receptors_in_repertoire_count: 100
                        generative_model:
                            chain: beta
                            default_model_name: humanTRB
                            model_path: null
                            type: OLGA
                        signals:
                            my_signal: 0.25
                            my_signal2: 0.01
                            my_signal__my_signal2: 0.02 # my_signal and my_signal2 will co-occur in 2% of the receptors in all 10 repertoires
                    sim_item2:
                        number_of_examples: 5
                        receptors_in_repertoire_count: 150
                        seed: 10 #
                        generative_model:
                            chain: beta
                            default_model_name: humanTRB
                            model_path: null
                            type: OLGA
                        signals:
                            my_signal: 0.75
                        default_clonal_frequency:
                            a: 2
                        sequence_len_limits:
                            min: 3


    """

    signal_proportions: Dict[Union[Signal, SignalPair], float]
    name: str = ""
    is_noise: bool = False
    seed: int = None
    generative_model: GenerativeModel = None
    number_of_examples: int = 0
    receptors_in_repertoire_count: int = 0
    false_positive_prob_in_receptors: float = 0.
    false_negative_prob_in_receptors: float = 0.
    immune_events: dict = field(default_factory=dict)
    default_clonal_frequency: dict = None
    sequence_len_limits: dict = None

    @property
    def signals(self) -> List[Signal]:
        return list(self.signal_proportions.keys())



import os
from pathlib import Path
from typing import List
import logging

import pandas as pd

from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.ml_methods.util.Util import Util as MLUtil
from immuneML.ml_metrics import ClusteringMetric
from immuneML.ml_metrics.ClusteringMetric import is_internal, is_external
from immuneML.presentation.TemplateParser import TemplateParser
from immuneML.presentation.html.Util import Util
from immuneML.reports.ReportResult import ReportResult
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.clustering.ClusteringState import ClusteringState


class ClusteringHTMLBuilder:
    CSS_PATH = EnvironmentSettings.html_templates_path / "css/custom.css"

    @staticmethod
    def build(state: ClusteringState) -> Path:
        base_path = PathBuilder.build(state.result_path / "../HTML_output/")
        html_map = ClusteringHTMLBuilder.make_html_map(state, base_path)
        result_file = base_path / f"Clustering_{state.config.name}.html"

        TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "Clustering.html",
                             template_map=html_map, result_path=result_file)

        # Generate split pages
        for split_id in range(state.config.split_config.split_count):
            ClusteringHTMLBuilder._make_split_page(state, split_id, base_path)

        # Generate detail pages for each setting in each split
        for split_id in range(state.config.split_config.split_count):
            for setting in state.config.clustering_settings:
                ClusteringHTMLBuilder._make_setting_details_page(state, split_id, setting, base_path)

        return result_file

    @staticmethod
    def make_html_map(state: ClusteringState, base_path: Path) -> dict:
        html_map = {
            "css_style": Util.get_css_content(ClusteringHTMLBuilder.CSS_PATH),
            "name": state.config.name,
            'immuneML_version': MLUtil.get_immuneML_version(),
            "full_specs": Util.get_full_specs_path(base_path),
            "logfile": Util.get_logfile_path(base_path),
            "clustering_reports": ClusteringHTMLBuilder._format_reports(state.clustering_report_results, base_path),
            "splits": [{"number": i + 1, "path": f"split_{i + 1}.html"} for i in range(state.config.split_config.split_count)],
            **Util.make_dataset_html_map(state.config.dataset)
        }
        return html_map

    @staticmethod
    def _make_split_page(state: ClusteringState, split_id: int, base_path: Path):
        split_map = {
            "css_style": Util.get_css_content(ClusteringHTMLBuilder.CSS_PATH),
            "name": state.config.name,
            "split_number": split_id + 1,
            "predictions": {
                "discovery": {
                    "path": os.path.relpath(state.predictions_paths[split_id]['discovery'], base_path),
                    "data": ClusteringHTMLBuilder._format_predictions_file(
                        state.predictions_paths[split_id]['discovery'])
                }
            },
            "performance": {
                "internal": {
                    "show": any(is_internal(m) for m in state.config.metrics),
                    "discovery": ClusteringHTMLBuilder._make_internal_performance_table(state, 'discovery', split_id)
                },
                "external": {
                    "show": any(is_external(m) for m in state.config.metrics),
                    "discovery": ClusteringHTMLBuilder._make_external_performance_tables(state, 'discovery', split_id)
                }
            },
            "setting_details": ClusteringHTMLBuilder._make_setting_links(state, split_id, base_path),
            "has_method_based": "method_based" in state.config.validation_type,
            "has_result_based": "result_based" in state.config.validation_type,
            "main_page_link": f"Clustering_{state.config.name}.html"
        }

        # Add validation results if present
        if "method_based" in state.config.validation_type:
            split_map["predictions"]["method_based"] = {
                "path": os.path.relpath(state.predictions_paths[split_id]['method_based_validation'], base_path),
                "data": ClusteringHTMLBuilder._format_predictions_file(
                    state.predictions_paths[split_id]['method_based_validation'])
            }
            split_map["performance"]["internal"]["method_based"] = ClusteringHTMLBuilder._make_internal_performance_table(
                state, 'method_based_validation', split_id)
            split_map["performance"]["external"]["method_based"] = ClusteringHTMLBuilder._make_external_performance_tables(
                state, 'method_based_validation', split_id)

        if "result_based" in state.config.validation_type:
            split_map["predictions"]["result_based"] = {
                "path": os.path.relpath(state.predictions_paths[split_id]['result_based_validation'], base_path),
                "data": ClusteringHTMLBuilder._format_predictions_file(
                    state.predictions_paths[split_id]['result_based_validation'])
            }
            split_map["performance"]["internal"]["result_based"] = ClusteringHTMLBuilder._make_internal_performance_table(
                state, 'result_based_validation', split_id)
            split_map["performance"]["external"]["result_based"] = ClusteringHTMLBuilder._make_external_performance_tables(
                state, 'result_based_validation', split_id)

        result_path = base_path / f"split_{split_id + 1}.html"
        TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "ClusteringSplit.html",
                             template_map=split_map, result_path=result_path)

    @staticmethod
    def _make_setting_links(state: ClusteringState, split_id: int, base_path: Path) -> List[dict]:
        return [{
            "name": setting.get_key(),
            "path": f"split_{split_id + 1}_{setting.get_key()}.html"
        } for setting in state.config.clustering_settings]

    @staticmethod
    def _make_setting_details_page(state: ClusteringState, split_id: int, setting, base_path: Path):
        template_map = {"css_style": Util.get_css_content(ClusteringHTMLBuilder.CSS_PATH), "split_number": split_id + 1,
                        "setting_name": setting.get_key(),
                        "has_method_based": "method_based" in state.config.validation_type,
                        "has_result_based": "result_based" in state.config.validation_type,
                        "discovery": ClusteringHTMLBuilder._get_analysis_results(
                            state, split_id, setting, "discovery", base_path)}

        # Add validation results if present
        if "method_based" in state.config.validation_type:
            template_map["method_based"] = ClusteringHTMLBuilder._get_analysis_results(
                state, split_id, setting, "method_based_validation", base_path)

        if "result_based" in state.config.validation_type:
            template_map["result_based"] = ClusteringHTMLBuilder._get_analysis_results(
                state, split_id, setting, "result_based_validation", base_path)

        result_path = base_path / f"split_{split_id + 1}_{setting.get_key()}.html"
        TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "ClusteringSettingDetails.html",
                             template_map=template_map, result_path=result_path)

    @staticmethod
    def _get_analysis_results(state: ClusteringState, split_id: int, setting, analysis_type: str,
                              base_path: Path) -> dict:
        cl_result = state.clustering_items[split_id]
        if hasattr(cl_result, analysis_type):
            analysis_result = getattr(cl_result, analysis_type)
            if analysis_result and setting.get_key() in analysis_result.items:
                item_result = analysis_result.items[setting.get_key()]
                return {
                    "predictions_path": os.path.relpath(state.predictions_paths[split_id][analysis_type], base_path),
                    "internal_performance": item_result.item.internal_performance.get_df().to_html(border=0,
                        justify='left', max_rows=None,
                        index=False) if item_result.item.internal_performance and any(ClusteringMetric.is_internal(metric) for metric in state.config.metrics) else None,
                    "external_performance": item_result.item.external_performance.get_df().to_html(border=0,
                        justify='left', max_rows=None,
                        index=False) if item_result.item.external_performance and any(ClusteringMetric.is_external(metric) for metric in state.config.metrics) else None,
                    "reports": ClusteringHTMLBuilder._format_reports(item_result.report_results, base_path)
                }
        return None

    @staticmethod
    def _format_predictions_file(file_path: Path) -> str:
        try:
            df = pd.read_csv(file_path)
            return df.to_html(border=0, classes="prediction-table", max_rows=None, justify='left', index=False)
        except:
            return "Error loading predictions"

    @staticmethod
    def _format_reports(reports: List[ReportResult], base_path: Path) -> dict:
        if not reports:
            return {"has_reports": False}

        formatted_reports = []
        for report in reports:
            if isinstance(report, ReportResult):
                formatted_report = {
                    "name": report.name,
                    "info": report.info if hasattr(report, "info") else None,
                    "output_figures": [],
                    "output_tables": [],
                    "output_text": []
                }

                # Process figures
                if hasattr(report, "output_figures"):
                    formatted_report["output_figures"] = [{
                        "name": fig.name,
                        "path": os.path.relpath(fig.path, base_path),
                        "is_embed": str(fig.path).endswith(('.html', '.svg'))
                    } for fig in report.output_figures]

                # Process tables
                if hasattr(report, "output_tables"):
                    for table in report.output_tables:
                        try:
                            df = pd.read_csv(table.path)
                            formatted_report["output_tables"].append({
                                "name": table.name,
                                "download_link": os.path.relpath(table.path, base_path),
                                "file_name": os.path.basename(table.path)
                            })
                        except Exception as e:
                            logging.warning(f"Error processing table {table.name}: {e}")

                # Process text outputs
                if hasattr(report, "output_text"):
                    formatted_report["output_text"] = [{
                        "name": text.name,
                        "download_link": os.path.relpath(text.path, base_path),
                        "file_name": os.path.basename(text.path),
                        "is_download_link": True
                    } for text in report.output_text]

                formatted_reports.append(formatted_report)

        return {
            "has_reports": True,
            "reports": formatted_reports
        }

    @staticmethod
    def _make_internal_performance_table(state: ClusteringState, analysis_type: str, split_id: int) -> str:
        cl_result = getattr(state.clustering_items[split_id], analysis_type, None)
        if not cl_result:
            return None

        performance_data = {
            "clustering setting": [],
            **{metric: [] for metric in state.config.metrics if is_internal(metric)}
        }

        for setting in state.config.clustering_settings:
            if setting.get_key() in cl_result.items:
                item = cl_result.items[setting.get_key()].item
                performance_data["clustering setting"].append(setting.get_key())
                for metric in state.config.metrics:
                    if is_internal(metric) and item.internal_performance:
                        value = item.internal_performance.get_df()[metric].values[0]
                        performance_data[metric].append(f"{value:.3f}")

        if performance_data["clustering setting"]:
            df = pd.DataFrame(performance_data)
            return df.to_html(border=0, justify='left', max_rows=None, index=False)
        return None

    @staticmethod
    def _make_external_performance_tables(state: ClusteringState, analysis_type: str, split_id: int) -> List[dict]:
        if not state.config.label_config:
            return []

        cl_result = getattr(state.clustering_items[split_id], analysis_type, None)
        if not cl_result:
            return []

        tables = []
        for label in state.config.label_config.get_labels_by_name():
            performance_data = {
                "clustering setting": [],
                **{metric: [] for metric in state.config.metrics if is_external(metric)}
            }

            for setting in state.config.clustering_settings:
                if setting.get_key() in cl_result.items:
                    item = cl_result.items[setting.get_key()].item
                    if item.external_performance:
                        performance_data["clustering setting"].append(setting.get_key())
                        for metric in state.config.metrics:
                            if is_external(metric):
                                value = item.external_performance.get_df().set_index(['metric']).loc[
                                    metric, label].item()
                                performance_data[metric].append(f"{value:.3f}")

            if performance_data["clustering setting"]:
                df = pd.DataFrame(performance_data)
                tables.append({
                    "label": label,
                    "performance_table": df.to_html(border=0, justify='left', max_rows=None, index=False)
                })

        return tables



import os
import shutil
import numpy as np
from unittest import TestCase

from immuneML.environment.Label import Label
from immuneML.caching.CacheType import CacheType
from immuneML.util.PathBuilder import PathBuilder
from immuneML.environment.Constants import Constants
from immuneML.data_model.encoded_data.EncodedData import EncodedData
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.ml_methods.classifiers.SillyClassifier import SillyClassifier


class TestSillyClassifier(TestCase):

    def setUp(self) -> None:
        os.environ[Constants.CACHE_TYPE] = CacheType.TEST.name

    def get_enc_data(self):
        # Creates a mock encoded data object with 8 random examples
        enc_data = EncodedData(examples=np.array([[1, 0, 0],
                                                  [0, 1, 1],
                                                  [1, 1, 1],
                                                  [0, 1, 1],
                                                  [1, 0, 0],
                                                  [0, 1, 1],
                                                  [1, 1, 1],
                                                  [0, 1, 1]]),
                               example_ids=list(range(8)),
                               feature_names=["a", "b", "c"],
                               labels={"my_label": ["yes", "no", "yes", "no", "yes", "no", "yes", "no"]},
                               encoding="random")

        label = Label(name="my_label", values=["yes", "no"], positive_class="yes")

        return enc_data, label

    def test_predictions(self):
        enc_data, label = self.get_enc_data()
        classifier = SillyClassifier(random_seed=50)

        # test fitting
        classifier.fit(enc_data, label)
        self.assertTrue(classifier.silly_model_fitted)

        # test 'predict'
        predictions = classifier.predict(enc_data, label)
        self.assertEqual(len(predictions[label.name]), len(enc_data.examples))

        # test 'predict_proba'
        prediction_probabilities = classifier.predict_proba(enc_data, label)
        self.assertEqual(len(prediction_probabilities[label.name][label.positive_class]), len(enc_data.examples))
        self.assertEqual(len(prediction_probabilities[label.name][label.get_binary_negative_class()]), len(enc_data.examples))
        self.assertTrue(all(0 <= pred <= 1 for pred in prediction_probabilities[label.name][label.positive_class]))
        self.assertTrue(all(0 <= pred <= 1 for pred in prediction_probabilities[label.name][label.get_binary_negative_class()]))

    def test_store_and_load(self):
        path = PathBuilder.build(EnvironmentSettings.tmp_test_path / "silly")
        enc_data, label = self.get_enc_data()
        classifier = SillyClassifier(random_seed=50)
        classifier.fit(enc_data, label)
        classifier.store(path)

        # Loading should be done in an 'empty' model (no parameters)
        classifier2 = SillyClassifier()
        classifier2.load(path)

        self.assertEqual(classifier.get_params(), classifier2.get_params())
        shutil.rmtree(path)




from dataclasses import dataclass

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.workflows.steps.StepParams import StepParams


@dataclass
class DataEncoderParams(StepParams):

    dataset: Dataset
    encoder: DatasetEncoder
    encoder_params: EncoderParams



from pathlib import Path

import numpy as np
import pandas as pd
import plotly.express as px

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class RandomDataPlot(DataReport):
    """
    This RandomDataPlot is a placeholder for a real Report.
    It plots some random numbers.

    **Specification arguments:**

    - n_points_to_plot (int): The number of random points to plot.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_report:
                    RandomDataPlot:
                        n_points_to_plot: 10

    """

    @classmethod
    def build_object(cls, **kwargs):
        # Here you may check the values of given user parameters
        # This will ensure immuneML will crash early (upon parsing the specification) if incorrect parameters are specified
        ParameterValidator.assert_type_and_value(kwargs['n_points_to_plot'], int, RandomDataPlot.__name__, 'n_points_to_plot', min_inclusive=1)

        return RandomDataPlot(**kwargs)

    def __init__(self, dataset: Dataset = None, result_path: Path = None, number_of_processes: int = 1, name: str = None,
                 n_points_to_plot: int = None):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
        self.n_points_to_plot = n_points_to_plot

    def check_prerequisites(self):
        # Here you may check properties of the dataset (e.g. dataset type), or parameter-dataset compatibility
        # and return False if the prerequisites are incorrect.
        # This will generate a user-friendly error message and ensure immuneML does not crash, but instead skips the report.
        # Note: parameters should be checked in 'build_object'
        return True

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        df = self._get_random_data()

        # utility function for writing a dataframe to a csv file
        # and creating a ReportOutput object containing the reference
        report_output_table = self._write_output_table(df, self.result_path / 'random_data.csv', name="Random data file")

        # Calling _safe_plot will internally call _plot, but ensure immuneML does not crash if errors occur
        report_output_fig = self._safe_plot(df=df)

        # Ensure output is either None or a list with item (not an empty list or list containing None)
        output_tables = None if report_output_table is None else [report_output_table]
        output_figures = None if report_output_fig is None else [report_output_fig]

        return ReportResult(name=self.name,
                            info="Some random numbers",
                            output_tables=output_tables,
                            output_figures=output_figures)

    def _get_random_data(self):
        return pd.DataFrame({"random_data_dim1": np.random.rand(self.n_points_to_plot),
                             "random_data_dim2": np.random.rand(self.n_points_to_plot)})

    def _plot(self, df: pd.DataFrame) -> ReportOutput:
        figure = px.scatter(df, x="random_data_dim1", y="random_data_dim2", template="plotly_white")
        figure.update_layout(template="plotly_white")

        file_path = self.result_path / "random_data.html"
        figure.write_html(str(file_path))
        return ReportOutput(path=file_path, name="Random data plot")




import logging
from pathlib import Path

import yaml
import os
import glob
import pandas as pd

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.app.ImmuneMLApp import ImmuneMLApp
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class Util:

    @staticmethod
    def check_parameters(yaml_path: Path, output_dir: Path, kwargs, location):
        assert isinstance(yaml_path, Path), f"{location}: yaml_path is {output_dir}, expected Path object."
        assert isinstance(output_dir, Path), f"{location}: output_dir is {output_dir}, expected Path object pointing to a folder to store the results."

        assert yaml_path.is_file(), f"{location}: path to the specification is not correct, got {yaml_path}, expecting path to a YAML file."


    @staticmethod
    def check_paths(specs: dict, tool_name: str):
        for key in specs.keys():
            if isinstance(specs[key], str):
                if "/" in specs[key] and specs[key] != "./" and any(name_part in key for name_part in ('path', 'file')):
                    logging.warning(f"{tool_name}: the paths in specification for Galaxy have to consist only of the filenames "
                                    f"as uploaded to Galaxy history beforehand. The problem occurs for the parameter {key}.")
            elif isinstance(specs[key], dict):
                Util.check_paths(specs[key], tool_name)

    @staticmethod
    def update_dataset_key(specs: dict, location, new_key="dataset"):
        dataset_keys = list(specs["definitions"]["datasets"].keys())
        assert len(dataset_keys) == 1, f"{location}: one dataset has to be defined under definitions/datasets, got {dataset_keys} instead."

        orig_key = dataset_keys[0]

        if orig_key != "dataset":
            specs["definitions"]["datasets"][new_key] = specs["definitions"]["datasets"][orig_key]
            specs["definitions"]["datasets"].pop(orig_key)

            for instruction_key in specs["instructions"].keys():
                if "dataset" in specs["instructions"][instruction_key]:
                    specs["instructions"][instruction_key]["dataset"] = new_key

                if "datasets" in specs["instructions"][instruction_key]:
                    specs["instructions"][instruction_key]["datasets"] = [new_key]

                if "analyses" in specs["instructions"][instruction_key]:
                    for analysis_key in specs["instructions"][instruction_key]["analyses"].keys():
                        specs["instructions"][instruction_key]["analyses"][analysis_key]["dataset"] = new_key

            logging.info(f"{location}: renamed dataset '{orig_key}' to '{new_key}'.")

    @staticmethod
    def update_result_paths(specs: dict, result_path: Path, yaml_path: Path):
        if 'datasets' in specs['definitions']:
            for key, item in specs["definitions"]["datasets"].items():
                if isinstance(item, dict) and 'params' in item.keys() and isinstance(item["params"], dict):
                    item['params']["result_path"] = str(result_path / key)
                    if item['format'] not in ['RandomRepertoireDataset', 'RandomReceptorDataset', 'RandomSequenceDataset']:
                        item['params']['path'] = str(yaml_path.parent)

        with yaml_path.open("w") as file:
            yaml.dump(specs, file)

    @staticmethod
    def check_instruction_type(specs: dict, tool_name, expected_instruction) -> str:
        ParameterValidator.assert_keys_present(list(specs.keys()), ['definitions', 'instructions'], tool_name, "YAML specification")
        assert len(list(specs['instructions'].keys())) == 1, f"{tool_name}: multiple instructions were given " \
                                                             f"({str(list(specs['instructions'].keys()))[1:-1]}), but only one instruction of type " \
                                                             f"{expected_instruction} should be specified."
        instruction_name = list(specs['instructions'].keys())[0]
        instruction_type = specs['instructions'][instruction_name]['type']
        assert instruction_type == expected_instruction, \
            f"{tool_name}: instruction type has to be '{expected_instruction}', got {instruction_type} instead."

        return instruction_name

    @staticmethod
    def check_export_format(specs: dict, tool_name: str, instruction_name: str):
        ParameterValidator.assert_keys_present(list(specs['instructions'][instruction_name].keys()), ["export_formats"], tool_name,
                                               f"{instruction_name}/export_formats")
        ParameterValidator.assert_type_and_value(specs['instructions'][instruction_name]["export_formats"], list, tool_name,
                                                 f"{instruction_name}/export_formats")

        assert len(specs['instructions'][instruction_name]["export_formats"]) == 1, \
            f"{tool_name}: only one format can be specified under export_formats parameter under " \
            f"{instruction_name}/export_formats, got {specs['instructions'][instruction_name]['export_formats']} instead."

        return specs['instructions'][instruction_name]["export_formats"][0]

    @staticmethod
    def run_tool(yaml_path, result_path):
        PathBuilder.build(result_path)
        app = ImmuneMLApp(yaml_path, result_path)
        app.run()

    @staticmethod
    def discover_dataset_path(dataset_name="dataset"):
        if os.path.exists(f"{dataset_name}.yaml"):
            dataset_path = f"{dataset_name}.yaml"
        else:
            discovered = glob.glob(f"*{dataset_name}*.yaml")

            if len(discovered) == 1:
                dataset_path = discovered[0]
            else:
                raise FileNotFoundError(f"Unable to locate '{dataset_name}.yaml'")

        return dataset_path


    @staticmethod
    def remove_path_from_filename(file_path):
        return str(Path(file_path).name)

    @staticmethod
    def reformat_galaxy_dataset(galaxy_dataset_path):
        dataset_yaml_file = galaxy_dataset_path / "dataset.yaml"
        assert dataset_yaml_file.is_file(), "Error: generated dataset.yaml not found"

        metadata_file = None

        with (dataset_yaml_file.open("r") as file):
            dataset_params = yaml.load(file, Loader=yaml.SafeLoader)

            if "metadata_file" in dataset_params:
                dataset_params["metadata_file"] = Util.remove_path_from_filename(dataset_params["metadata_file"])
                metadata_file = galaxy_dataset_path / dataset_params["metadata_file"]

            if "filename" in dataset_params:
                dataset_params["filename"] = str(Path(dataset_params["filename"]).name)

        with dataset_yaml_file.open("w") as file:
            yaml.dump(dataset_params, file)

        if metadata_file is not None:
            metadata_content = pd.read_csv(metadata_file, sep=",")
            metadata_content["filename"] = [Util.remove_path_from_filename(filename) for filename in metadata_content["filename"]]
            metadata_content.to_csv(metadata_file, index=None)

    @staticmethod
    def export_galaxy_dataset(dataset, result_path):
        try:
            PathBuilder.build(result_path / 'galaxy_dataset')
            AIRRExporter.export(dataset, result_path / "galaxy_dataset/")
            dataset_file = list(glob.glob(str(result_path / "galaxy_dataset/*.yaml")))[0]
            os.rename(dataset_file, result_path / "galaxy_dataset/dataset.yaml")
            Util.reformat_galaxy_dataset(result_path / "galaxy_dataset/")
        except Exception as e:
            raise RuntimeError(f"Error when exporting Galaxy dataset: {e}.")


from pathlib import Path

from immuneML.IO.dataset_export.DataExporter import DataExporter
from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReflectionHandler import ReflectionHandler
from immuneML.workflows.instructions.subsampling.SubsamplingInstruction import SubsamplingInstruction


class SubsamplingParser:

    def parse(self, key: str, instruction: dict, symbol_table: SymbolTable, path: Path = None) -> SubsamplingInstruction:
        valid_keys = ["type", "dataset", "subsampled_dataset_sizes", "subsampled_repertoire_size"]
        ParameterValidator.assert_keys(instruction.keys(), valid_keys, SubsamplingParser.__name__, key)

        dataset_keys = symbol_table.get_keys_by_type(SymbolType.DATASET)
        ParameterValidator.assert_in_valid_list(instruction['dataset'], dataset_keys, SubsamplingParser.__name__, f'{key}/dataset')

        dataset = symbol_table.get(instruction['dataset'])
        ParameterValidator.assert_type_and_value(instruction['subsampled_dataset_sizes'], list, SubsamplingParser.__name__, f'{key}/subsampled_dataset_sizes')
        ParameterValidator.assert_all_type_and_value(instruction['subsampled_dataset_sizes'], int, SubsamplingParser.__name__,
                                                     f'{key}/subsampled_dataset_sizes', 1, dataset.get_example_count())

        ParameterValidator.assert_type_and_value(instruction['subsampled_repertoire_size'], int,
                                                 SubsamplingParser.__name__, f'{key}/subsampled_repertoire_size',
                                                 nullable=True)

        return SubsamplingInstruction(dataset=dataset,
                                      subsampled_repertoire_size=instruction['subsampled_repertoire_size'],
                                      subsampled_dataset_sizes=instruction['subsampled_dataset_sizes'],
                                      name=key)



import numpy as np

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams


class ShannonDiversityEncoder(DatasetEncoder):
    """
    ShannonDiversity encoder calculates the Shannon diversity index for each repertoire in a dataset. The diversity is
    computed as:

    .. math::

        diversity = - \\sum_{i=1}^{n} p_i \\log(p_i)

    where :math:`p_i` is the clonal count for each unique sequence in the repertoire (from duplicate_count field)
    divided by the total clonal counts, and :math:`n` is the total number of clonotypes (sequences) in the repertoire.


    **Dataset type:**

    - RepertoireDataset

    **Specification arguments:**

    No arguments are needed for this encoder.

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                shannon_div_enc: ShannonDiversity

    """

    def __init__(self, name: str = None):
        super().__init__(name=name)

    @staticmethod
    def build_object(dataset: Dataset, **params):
        assert isinstance(dataset, RepertoireDataset), \
            f"{ShannonDiversityEncoder.__name__}: Dataset must be of type RepertoireDataset, but got {type(dataset)}."
        return ShannonDiversityEncoder(**params)

    def encode(self, dataset, params: EncoderParams) -> Dataset:
        assert isinstance(dataset, RepertoireDataset), \
            f"{ShannonDiversityEncoder.__name__}: Dataset must be of type RepertoireDataset, but got {type(dataset)}."

        examples = CacheHandler.memo_by_params((dataset.identifier, ShannonDiversityEncoder.__name__,
                                                params.label_config.get_labels_by_name() if params.encode_labels else ''),
                                               lambda: self._encode(dataset, params))

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(examples=examples,
                                                   labels=dataset.get_metadata(params.label_config.get_labels_by_name()) if params.encode_labels else {},
                                                   example_ids=dataset.get_example_ids(),
                                                   encoding=ShannonDiversityEncoder.__name__)
        return encoded_dataset

    def _encode(self, dataset: RepertoireDataset, params: EncoderParams) -> np.ndarray:
        entropies = []
        for repertoire in dataset.repertoires:
            data = repertoire.data
            probabilities = data.duplicate_count / np.sum(data.duplicate_count)
            entropy = -1. * np.sum(probabilities * np.log(probabilities))
            entropies.append(entropy)
        return np.array(entropies).reshape(-1, 1)



import shutil
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd

from immuneML.data_model.SequenceParams import Chain
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.bnp_util import write_yaml, read_yaml, get_sequence_field_name
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.util.PathBuilder import PathBuilder


class PWM(GenerativeModel):
    """

    This is a baseline implementation of a positional weight matrix. It is estimated from a set of sequences for each
    of the different lengths that appear in the dataset.


    **Specification arguments:**

    - locus (str): which chain is generated (for now, it is only assigned to the generated sequences)

    - sequence_type (str): amino_acid or nucleotide

    - region_type (str): which region type to use (e.g., IMGT_CDR3), this is only assigned to the generated sequences


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_pwm:
                    PWM:
                        locus: beta
                        sequence_type: amino_acid
                        region_type: IMGT_CDR3

    """

    @classmethod
    def load_model(cls, path: Path):

        assert path.exists(), f"{cls.__name__}: {path} does not exist."

        model_overview_file = path / 'model_overview.yaml'
        length_probs_file = path / "length_probabilities.yaml"

        for file in [model_overview_file, length_probs_file]:
            assert file.exists(), f"{cls.__name__}: {file} is not a file."

        length_probs = read_yaml(length_probs_file)
        model_overview = read_yaml(model_overview_file)
        pwm_matrix = {}

        pwm = PWM(locus=model_overview['locus'], sequence_type=model_overview['sequence_type'],
                  region_type=model_overview['region_type'])

        for file, length in [(path / f'pwm_len_{length}.csv', length) for length in length_probs.keys()]:
            assert file.exists(), f"{cls.__name__}: {file} is not a file."
            pwm_matrix[length] = pd.read_csv(str(file))

            assert pwm_matrix[length].iloc[:, 0].tolist() == EnvironmentSettings.get_sequence_alphabet(
                pwm.sequence_type), \
                (f"{cls.__name__}: the row names in the PWM for length {length} don't match the expected row names.\n"
                 f"Obtained:\n{pwm_matrix[length].index.tolist()}\nExpected:"
                 f"\n{EnvironmentSettings.get_sequence_alphabet(pwm.sequence_type)}")

            pwm_matrix[length] = pwm_matrix[length].iloc[:, 1:].values

            expected_shape = (len(EnvironmentSettings.get_sequence_alphabet(pwm.sequence_type)), length)
            assert pwm_matrix[length].shape == expected_shape, \
                (f"{cls.__name__}: PWM matrix for length {length} has shape {pwm_matrix[length].shape}, "
                 f"but expected {expected_shape}.")

        pwm.length_probs = length_probs
        pwm.pwm_matrix = pwm_matrix
        return pwm

    def __init__(self, locus, sequence_type: str, region_type: str, name: str = None):
        super().__init__(locus, name=name, region_type=RegionType[region_type.upper()])
        self.sequence_type = SequenceType[sequence_type.upper()]
        self.pwm_matrix = None
        self.length_probs = None

    def fit(self, data: SequenceDataset, path: Path = None):
        self.set_locus(data)
        sequences = data.get_attribute(get_sequence_field_name(self.region_type, self.sequence_type))
        lengths, counts = np.unique(sequences.lengths, return_counts=True)

        self.length_probs = dict(zip(lengths, counts))
        self.length_probs = {int(length): float(count / sum(counts)) for length, count in self.length_probs.items()}
        self.pwm_matrix = {length: None for length in lengths}

        alphabet = EnvironmentSettings.get_sequence_alphabet(self.sequence_type)

        for length in lengths:
            seq_subset = sequences[sequences.lengths == length]
            self.pwm_matrix[length] = np.zeros((len(alphabet), length))
            for position in range(length):
                for i, letter in enumerate(alphabet):
                    self.pwm_matrix[length][i, position] = np.sum([1 for seq in seq_subset if seq[position] == letter])

            # Normalize counts to obtain probabilities
            total_counts = np.sum(self.pwm_matrix[length], axis=0)
            self.pwm_matrix[length] = self.pwm_matrix[length] / total_counts

    def is_same(self, model) -> bool:
        raise NotImplementedError

    def generate_sequences(self, count: int, seed: int, path: Path, sequence_type: SequenceType, compute_p_gen: bool):
        np.random.seed(seed)
        assert sequence_type == self.sequence_type

        sequences = []

        for _ in range(count):
            sequence_length = np.random.choice(list(self.length_probs.keys()), p=list(self.length_probs.values()))
            sequence = "".join(np.random.choice(EnvironmentSettings.get_sequence_alphabet(self.sequence_type),
                                                p=self.pwm_matrix[sequence_length][:, i])
                               for i in range(sequence_length))

            sequences.append(sequence)

        dataset = self._export_gen_dataset(sequences, path)

        return dataset

    def _export_gen_dataset(self, sequences: List[str], path: Path) -> SequenceDataset:
        count = len(sequences)
        df = pd.DataFrame({get_sequence_field_name(self.region_type, self.sequence_type): sequences,
                           'locus': [self.locus.to_string() for _ in range(count)],
                           'gen_model_name': [self.name for _ in range(count)]})

        return SequenceDataset.build_from_partial_df(df, PathBuilder.build(path), 'synthetic_dataset',
                                                     {'gen_model_name': [self.name]}, {'gen_model_name': str})

    def compute_p_gens(self, sequences, sequence_type: SequenceType) -> np.ndarray:
        raise NotImplementedError

    def compute_p_gen(self, sequence: dict, sequence_type: SequenceType) -> float:
        raise NotImplementedError

    def can_compute_p_gens(self) -> bool:
        return True

    def can_generate_from_skewed_gene_models(self) -> bool:
        return False

    def generate_from_skewed_gene_models(self, v_genes: list, j_genes: list, seed: int, path: Path,
                                         sequence_type: SequenceType, batch_size: int, compute_p_gen: bool):
        raise RuntimeError

    def save_model(self, path: Path) -> Path:
        model_path = PathBuilder.build(path / 'model')
        write_yaml(yaml_dict=self.length_probs, filename=model_path / 'length_probabilities.yaml')
        for length in self.pwm_matrix:
            (pd.DataFrame(data=self.pwm_matrix[length], columns=list(range(length)),
                          index=EnvironmentSettings.get_sequence_alphabet(self.sequence_type))
             .to_csv(model_path / f'pwm_len_{length}.csv'))

        write_yaml(filename=model_path / 'model_overview.yaml',
                   yaml_dict={'type': 'PWM', 'locus': self.locus.name, 'sequence_type': self.sequence_type.name,
                              'region_type': self.region_type.name})

        return Path(shutil.make_archive(str(path / f'trained_model_{self.name}'), "zip", str(path / 'model'))).absolute()



import copy
import logging
from collections import ChainMap
from datetime import datetime
import logging
import uuid
from pathlib import Path
from uuid import uuid4

import pandas as pd

from immuneML import Constants
from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.data_model.bnp_util import write_yaml, write_dataset_yaml
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class RepertoireDataset(Dataset):

    @classmethod
    def build_from_objects(cls, **kwargs):
        ParameterValidator.assert_keys_present(list(kwargs.keys()), ['repertoires', 'path'], RepertoireDataset.__name__,
                                               RepertoireDataset.__name__)
        ParameterValidator.assert_all_type_and_value(kwargs['repertoires'], Repertoire, RepertoireDataset.__name__,
                                                     'repertoires')

        assert len(kwargs['repertoires']) > 0, "Cannot to construct a repertoire dataset without repertories."

        metadata_df = pd.DataFrame.from_records(
            [{**rep.metadata, **{'filename': rep.data_filename.name, 'identifier': rep.identifier}}
             for rep in kwargs['repertoires']])

        metadata_df.drop(columns=['field_list', 'type_dict_dynamic_fields'], inplace=True, errors='ignore')

        metadata_path = PathBuilder.build(kwargs['path']) / 'metadata.csv'
        metadata_df.to_csv(metadata_path, index=False)

        logging.info(f"Made new repertoire dataset at {kwargs['path']} with metadata at {metadata_path} "
                     f"with {len(kwargs['repertoires'])} repertoires.")

        name = kwargs.get('name', None)

        dataset = RepertoireDataset(repertoires=kwargs['repertoires'], metadata_file=metadata_path, name=name)
        label_names = list(dataset.get_label_names(refresh=True))
        dataset.labels = {label: list(set(values)) for label, values in dataset.get_metadata(label_names).items()}

        dataset_file = PathBuilder.build(kwargs['path']) / 'dataset.yaml'
        dataset_meta_content = cls.create_metadata_dict(labels=dataset.labels,
                                                        identifier=dataset.identifier,
                                                        name=dataset.name,
                                                        metadata_file=str(metadata_path.name))

        write_dataset_yaml(dataset_file, dataset_meta_content)
        dataset.dataset_file = dataset_file

        return dataset

    @classmethod
    def build(cls, **kwargs):
        ParameterValidator.assert_keys_present(list(kwargs.keys()),
                                               ['metadata_file', 'name', 'repertoire_ids', 'metadata_fields'],
                                               RepertoireDataset.__name__, "repertoire dataset")
        repertoires = []
        metadata_df = pd.read_csv(kwargs['metadata_file'], comment=Constants.COMMENT_SIGN)
        for index, row in metadata_df.iterrows():
            filename = Path(kwargs['metadata_file']).parent / row['filename']
            if not filename.is_file() and 'repertoires' in str(filename):
                filename = filename.parent.parent / Path(row['filename']).name
            repertoire = Repertoire(data_filename=filename,
                                    metadata_filename=filename.parent / f'{filename.stem}_metadata.yaml',
                                    identifier=row['identifier'] if 'identifier' in row else uuid4().hex)
            repertoires.append(repertoire)

        if "repertoire_id" in kwargs.keys() and "repertoires" not in kwargs.keys() and kwargs[
            'repertoire_id'] is not None:
            assert all(rep.identifier == kwargs['repertoire_id'][i] for i, rep in enumerate(repertoires)), \
                f"{RepertoireDataset.__name__}: repertoire ids from the dataset file and metadata file don't match for the dataset " \
                f"{kwargs['name']} with identifier {kwargs['identifier']}."

        return RepertoireDataset(**{**kwargs, **{"repertoires": repertoires}})

    @classmethod
    def create_metadata_dict(cls, metadata_file, labels, name, identifier=None):
         return {"metadata_file": Path(metadata_file).name,
                 # "type_dict_dynamic_fields": type_dict,
                 "labels": {} if labels is None else labels,
                 "name": name,
                 "identifier": identifier if identifier is not None else uuid4().hex,
                 "dataset_type": cls.__name__,
                 "timestamp": datetime.now()}

    def __init__(self, labels: dict = None, encoded_data: EncodedData = None, repertoires: list = None,
                 identifier: str = None, metadata_file: Path = None, name: str = None, metadata_fields: list = None,
                 repertoire_ids: list = None, dataset_file: Path = None):
        super().__init__(encoded_data=encoded_data, name=name,
                         identifier=identifier if identifier is not None else uuid4().hex, labels=labels)
        self.metadata_file = Path(metadata_file) if metadata_file is not None else None
        self.metadata_fields = metadata_fields
        self.repertoire_ids = repertoire_ids
        self.repertoires = repertoires
        self.dataset_file = dataset_file

    def clone(self, keep_identifier: bool = False):
        dataset = RepertoireDataset(self.labels, copy.deepcopy(self.encoded_data), copy.deepcopy(self.repertoires),
                                    metadata_file=self.metadata_file, name=self.name)
        if keep_identifier:
            dataset.identifier = self.identifier
        return dataset

    def add_encoded_data(self, encoded_data: EncodedData):
        self.encoded_data = encoded_data

    def get_data(self, batch_size: int = 1):
        return self.repertoires

    def get_locus(self):
        loci = set()

        for repertoire in self.get_data():
            loci.update(repertoire.get_locus())

        return sorted(loci)

    def get_repertoire(self, index: int = -1, repertoire_identifier: str = "") -> Repertoire:
        assert index != -1 or repertoire_identifier != "", \
            "RepertoireDataset: cannot get repertoire since the index nor identifier are set."
        return self.repertoires[index] if index != -1 else \
            [rep for rep in self.repertoires if rep.identifier == repertoire_identifier][0]

    def get_example_count(self):
        return len(self.repertoires)

    def get_metadata_fields(self, refresh=False):
        """Returns the list of metadata fields, includes also the fields that will typically not be used as labels,
        like filename or identifier"""
        if self.metadata_fields is None or refresh:
            df = pd.read_csv(self.metadata_file, sep=",", nrows=0, comment=Constants.COMMENT_SIGN)
            self.metadata_fields = df.columns.values.tolist()
        return self.metadata_fields

    def get_label_names(self, refresh=False):
        """Returns the list of metadata fields which can be used as labels; if refresh=True, it reloads the fields
        from disk"""
        all_metadata_fields = set(self.get_metadata_fields(refresh))
        for non_label in ["subject_id", "filename", "repertoire_id", "identifier", "type_dict_dynamic_fields", "dataset_type"]:
            if non_label in all_metadata_fields:
                all_metadata_fields.remove(non_label)

        return all_metadata_fields

    def get_metadata(self, field_names: list, return_df: bool = False):
        assert isinstance(self.metadata_file, Path) and self.metadata_file.is_file(), \
            (f"RepertoireDataset: for dataset {self.name} (id: {self.identifier}) metadata file is not set properly. "
             f"The metadata file points to {self.metadata_file}.")

        try:
            df = pd.read_csv(self.metadata_file, sep=",", usecols=field_names, comment=Constants.COMMENT_SIGN)
        except ValueError:
            df = pd.read_csv(self.metadata_file, sep=",", comment=Constants.COMMENT_SIGN)
            df.drop(columns=[col for col in df.columns if col not in field_names], inplace=True)
            logging.warning(f"RepertoireDataset: metadata file {self.metadata_file} does not contain the "
                            f"following fields: {[col for col in field_names if col not in df.columns]}. ")

        if return_df:
            return df
        else:
            return df.to_dict("list")

    def get_filenames(self):
        return [Path(filename) for filename in self.get_metadata(["filename"])["filename"]]

    def _build_new_metadata(self, indices, path: Path) -> Path:
        if self.metadata_file:
            df = pd.read_csv(self.metadata_file, comment=Constants.COMMENT_SIGN)
            df = df.iloc[indices, :]
            df.to_csv(path, index=False)
            return path
        else:
            return None

    def make_subset(self, example_indices, path: Path, dataset_type: str):
        """
        Creates a new dataset object with only those examples (repertoires) available which were given by index in example_indices argument.

        Args:
            example_indices (list): a list of indices of examples (repertoires) to use in the new dataset
            path (Path): a path where to store the newly created dataset
            dataset_type (str): a type of the dataset used as a part of the name of the resulting dataset; the values are defined as constants in :py:obj:`~immuneML.data_model.dataset.Dataset.Dataset`

        Returns:

            a new RepertoireDataset object which includes only the repertoires specified under example_indices

        """

        metadata_file = self._build_new_metadata(example_indices, path / f"{dataset_type}_metadata.csv")
        new_dataset = RepertoireDataset(repertoires=[self.repertoires[i] for i in example_indices],
                                        labels=copy.deepcopy(self.labels),
                                        metadata_file=metadata_file, identifier=str(uuid4()))

        return new_dataset

    def get_repertoire_ids(self) -> list:
        """Returns a list of repertoire identifiers, same as get_example_ids()"""
        if self.repertoire_ids is None:
            self.repertoire_ids = [str(repertoire.identifier) for repertoire in self.repertoires]
        return self.repertoire_ids

    def get_example_ids(self):
        """Returns a list of example identifiers"""
        return self.get_repertoire_ids()

    def get_subject_ids(self):
        """Returns a list of subject identifiers"""
        return self.get_metadata(["subject_id"])["subject_id"]

    def get_data_from_index_range(self, start_index: int, end_index: int):
        return self.repertoires[start_index:end_index + 1]



from pathlib import Path
from typing import List

import pandas as pd
import plotly.graph_objects as go

from immuneML.ml_metrics import ClusteringMetric
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.clustering_reports.ClusteringReport import ClusteringReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.clustering.ClusteringState import ClusteringState, ClusteringResultPerRun


class ExternalLabelMetricHeatmap(ClusteringReport):
    """
    This report creates heatmaps comparing clustering methods against external labels for each metric.
    For each external label and metric combination, it creates:

    1. A table showing the metric values for each combination of clustering method and external label

    2. A heatmap visualization of these values

    The external labels and metrics are automatically determined from the clustering instruction specification.

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        reports:
            my_external_label_metric_heatmap: ExternalLabelMetricHeatmap

    """

    @classmethod
    def build_object(cls, **kwargs):
        ParameterValidator.assert_keys(list(kwargs.keys()), ['name'],
                                     ExternalLabelMetricHeatmap.__name__, ExternalLabelMetricHeatmap.__name__)
        return ExternalLabelMetricHeatmap(**kwargs)

    def __init__(self, name: str = None, state: ClusteringState = None,
                 result_path: Path = None, number_of_processes: int = 1):
        super().__init__(name, result_path, number_of_processes, state)
        self.desc = "External Label - Clustering Heatmap"

    def _generate(self) -> ReportResult:
        self.result_path = PathBuilder.build(self.result_path / self.name)
        report_outputs = []
        external_labels = self.state.config.label_config.get_labels_by_name()

        if not external_labels or len(external_labels) == 0:
            return ReportResult(
                name=f"{self.desc} ({self.name})",
                info="No external labels were found in the clustering state's label configuration."
            )

        metrics = [metric for metric in self.state.config.metrics if ClusteringMetric.is_external(metric)]

        for metric in metrics:
            # For each split in the clustering results
            for split_idx, clustering_results in enumerate(self.state.clustering_items):
                # Process discovery results
                if clustering_results.discovery:
                    report_outputs.extend(self._process_analysis_results(
                        clustering_results.discovery,
                        f"discovery_split_{split_idx + 1}",
                        metric
                    ))

                # Process method-based validation results if available
                if clustering_results.method_based_validation:
                    report_outputs.extend(self._process_analysis_results(
                        clustering_results.method_based_validation,
                        f"method_based_validation_split_{split_idx + 1}",
                        metric
                    ))

                # Process result-based validation results if available
                if clustering_results.result_based_validation:
                    report_outputs.extend(self._process_analysis_results(
                        clustering_results.result_based_validation,
                        f"result_based_validation_split_{split_idx + 1}",
                        metric
                    ))

        if not report_outputs:
            return ReportResult(
                name=f"{self.desc} ({self.name})",
                info="No results were generated. This could be because no metrics were computed."
            )

        return ReportResult(
            name=f"{self.desc} ({self.name})",
            info="Heatmaps of metric values for clustering methods versus external labels",
            output_tables=[output for output in report_outputs if output.path.suffix == '.csv'],
            output_figures=[output for output in report_outputs if output.path.suffix in ['.html', '.png']],
        )

    def _process_analysis_results(self, analysis_results: ClusteringResultPerRun, analysis_name: str,
                                  metric: ClusteringMetric) -> List[ReportOutput]:
        outputs = []

        external_labels = self.state.config.label_config.get_labels_by_name()

        # Create a DataFrame with metric values
        df = pd.DataFrame(0, index=list(analysis_results.items.keys()), columns=external_labels)

        for setting_key, item_result in analysis_results.items.items():
            performance_df = item_result.item.external_performance.get_df()
            for label in external_labels:
                df.loc[setting_key, label] = performance_df[performance_df['metric'] == metric][label].values[0]

        # Create heatmap
        fig = go.Figure(data=go.Heatmap(
            z=df.values,
            x=df.columns,  # label values
            y=df.index,    # clustering methodsa
            colorscale='Darkmint',
            text=df.values,
            texttemplate='%{text:.3f}',
            hovertemplate=metric + ': %{z:.3f}<br>external label: %{x}<br>clustering setting: %{y}<extra></extra>',
            textfont={"size": 15},
            hoverongaps=False
        ))

        fig.update_layout(
            template='plotly_white'
        )

        # Save heatmap
        heatmap_path = self.result_path / f"{analysis_name}_{metric}_heatmap.html"
        plot_path = PlotlyUtil.write_image_to_file(fig, heatmap_path, df.shape[0])

        outputs.append(ReportOutput(
            path=plot_path,
            name=f"Heatmap for {metric.replace('_', ' ')} ({analysis_name.replace('_', ' ')})"
        ))

        # Save metric table
        table_path = self.result_path / f"{analysis_name}_{metric}.csv"
        df.reset_index().rename(columns={'index': 'clustering_setting'}).to_csv(table_path, index=False)
        outputs.append(ReportOutput(
            path=table_path,
            name=f"Metric values for {metric} ({analysis_name.replace('_', ' ')})"
        ))

        return outputs

    def check_prerequisites(self):
        if not self.state:
            return False

        if not self.state.clustering_items:
            return False

        if not self.state.config.label_config:
            return False

        return True



import os
import shutil
import unittest

from immuneML.caching.CacheType import CacheType
from immuneML.data_model.EncodedData import EncodedData
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.environment.Constants import Constants
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.encodings.silly.SillyEncoder import SillyEncoder
from immuneML.environment.Label import Label
from immuneML.simulation.dataset_generation.RandomDatasetGenerator import RandomDatasetGenerator


class TestSillyEncoder(unittest.TestCase):

    def setUp(self) -> None:
        os.environ[Constants.CACHE_TYPE] = CacheType.TEST.name

    def _get_mock_sequence_dataset(self, path):
        # Create a mock SequenceDataset with 10 sequences of length 15,
        # and a label called 'binding' with 50% chance of having status 'yes' or 'no'
        dataset = RandomDatasetGenerator.generate_sequence_dataset(sequence_count=10,
                                                                   length_probabilities={15: 1},
                                                                   labels={"binding": {"yes": 0.5, "no": 0.5}},
                                                                   path=path)

        label_config = LabelConfiguration(labels=[Label(name="binding", values=["yes", "no"])])

        return dataset, label_config

    def _get_mock_receptor_dataset(self, path):
        # Create a mock ReceptorDataset with 10 receptors with sequences of length 15,
        # and a label called 'binding' with 50% chance of having status 'yes' or 'no'
        dataset = RandomDatasetGenerator.generate_receptor_dataset(receptor_count=10,
                                                                   chain_1_length_probabilities={15: 1},
                                                                   chain_2_length_probabilities={15: 1},
                                                                   labels={"binding": {"yes": 0.5, "no": 0.5}},
                                                                   path=path)

        label_config = LabelConfiguration(labels=[Label(name="binding", values=["yes", "no"])])

        return dataset, label_config

    def _get_mock_repertoire_dataset(self, path):
        # Create a mock RepertoireDataset with 10 repertoires, each containing 50 sequences of length 15,
        # and a label called 'disease' with 50% chance of having status 'yes' or 'no'
        dataset = RandomDatasetGenerator.generate_repertoire_dataset(repertoire_count=10,
                                                           sequence_count_probabilities={50: 1},
                                                           sequence_length_probabilities={15: 1},
                                                           labels={"disease": {"yes": 0.5, "no": 0.5}},
                                                           path=path)

        label_config = LabelConfiguration(labels=[Label(name="disease", values=["yes", "no"])])

        return dataset, label_config

    def test_silly_sequence_encoder(self):
        tmp_path = EnvironmentSettings.tmp_test_path / "silly_sequence/"
        sequence_dataset, label_config = self._get_mock_sequence_dataset(tmp_path)
        self._test_silly_encoder(tmp_path, sequence_dataset, label_config)

    def test_silly_receptor_encoder(self):
        tmp_path = EnvironmentSettings.tmp_test_path / "silly_receptor/"
        receptor_dataset, label_config = self._get_mock_receptor_dataset(tmp_path)
        self._test_silly_encoder(tmp_path, receptor_dataset, label_config)

    def test_silly_repertoire_encoder(self):
        tmp_path = EnvironmentSettings.tmp_test_path / "silly_repertoire/"
        receptor_dataset, label_config = self._get_mock_repertoire_dataset(tmp_path)
        self._test_silly_encoder(tmp_path, receptor_dataset, label_config)

    def _test_silly_encoder(self, tmp_path, dataset, label_config):
        # test getting a SillyEncoder from the build_object method
        params = {"random_seed": 1, "embedding_len": 3}
        encoder = SillyEncoder.build_object(dataset, **params)
        self.assertIsInstance(encoder, SillyEncoder)

        # test encoding data
        encoded_dataset = encoder.encode(dataset,
                                         params=EncoderParams(result_path=tmp_path,
                                                              label_config=label_config))

        # the result must be a Dataset (of the same subtype as the original dataset) with EncodedData attached
        self.assertIsInstance(encoded_dataset, dataset.__class__)
        self.assertIsInstance(encoded_dataset.encoded_data, EncodedData)

        # testing the validity of the encoded data
        self.assertEqual(dataset.get_example_ids(), encoded_dataset.encoded_data.example_ids)
        self.assertTrue((encoded_dataset.encoded_data.examples >= 0).all())
        self.assertTrue((encoded_dataset.encoded_data.examples <= 1).all())

        # don't forget to remove the temporary data
        shutil.rmtree(tmp_path)



from collections import Counter
from typing import List

import pandas as pd
from sklearn.preprocessing import StandardScaler

from immuneML.analysis.data_manipulation.NormalizationType import NormalizationType
from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.preprocessing.FeatureScaler import FeatureScaler
from immuneML.util.NumpyHelper import NumpyHelper
from immuneML.util.ParameterValidator import ParameterValidator


class GeneFrequencyEncoder(DatasetEncoder):
    """
    GeneFrequencyEncoder represents a repertoire by the frequency of V and/or J genes used.

    **Dataset type:**
    - RepertoireDatasets

    **Specification arguments:**

    - genes (list): List of genes to use for the encoding. Possible values are 'V', and 'J'. At least one gene must be
      specified.

    - normalization_type (str): Type of normalization to apply to the gene frequencies. Possible values are 'none',
      'binary', 'relative_frequency', 'max', 'l2'. Defaults to 'relative_frequency'.

    - scale_to_zero_mean (bool): Whether to scale the features to zero mean. Defaults to True.

    - scale_to_unit_variance (bool): Whether to scale the features to unit variance. Defaults to True.

    **YAML specification:**

    .. code-block:: yaml

        encodings:
            gene_frequency_encoding:
                GeneFrequency:
                    genes: [V, J]
                    normalization_type: relative_frequency
                    scale_to_unit_variance: true
                    scale_to_zero_mean: true

    """

    def __init__(self, genes: List[str], normalization_type: NormalizationType, scale_to_zero_mean: bool,
                 scale_to_unit_variance: bool, name: str = None):
        super().__init__(name=name)
        self.genes = genes
        self.normalization_type = normalization_type
        self.scale_to_zero_mean = scale_to_zero_mean
        self.scale_to_unit_variance = scale_to_unit_variance
        self.scaler = None
        self.feature_names = None

    @property
    def all_feature_names(self) -> List[str]:
        return [feature_name for gene in self.genes for feature_name in self.feature_names[gene]]

    @staticmethod
    def build_object(dataset: Dataset, **params):
        valid_keys = ['genes', 'normalization_type', 'scale_to_zero_mean', 'scale_to_unit_variance', 'name']
        ParameterValidator.assert_keys(params.keys(), valid_keys, "GeneFrequencyEncoder", "parameters", exclusive=False)
        ParameterValidator.assert_all_in_valid_list(params['genes'], ['V', 'J'], "GeneFrequencyEncoder", "genes")
        ParameterValidator.assert_type_and_value(params['scale_to_zero_mean'], bool, "GeneFrequencyEncoder",
                                                 "scale_to_zero_mean")
        ParameterValidator.assert_type_and_value(params['scale_to_unit_variance'], bool, "GeneFrequencyEncoder",
                                                 "scale_to_unit_variance")

        normalization_type = NormalizationType[params['normalization_type'].upper()]

        return GeneFrequencyEncoder(genes=params['genes'], normalization_type=normalization_type,
                                    scale_to_zero_mean=params['scale_to_zero_mean'],
                                    scale_to_unit_variance=params['scale_to_unit_variance'],
                                    name=params.get('name', 'gene_frequency'))

    def encode(self, dataset, params: EncoderParams) -> Dataset:
        if isinstance(dataset, RepertoireDataset):
            return self._encode_repertoire_dataset(dataset, params)
        else:
            raise RuntimeError(f"{self.__class__.__name__}: {self.name}: invalid dataset type: {type(dataset)}.")

    def _encode_repertoire_dataset(self, dataset: RepertoireDataset, params: EncoderParams) -> RepertoireDataset:
        counters = {gene: [] for gene in self.genes}
        for rep in dataset.repertoires:
            for gene in self.genes:
                genes = [gene_call.split("*")[0] for gene_call in getattr(rep.data, f'{gene.lower()}_call').tolist()]
                counters[gene].append(Counter(genes))

        gene_dfs = {gene: pd.DataFrame(counters[gene]).fillna(0) for gene in self.genes}

        if params.learn_model:
            self.feature_names = {gene: df.columns.tolist() for gene, df in gene_dfs.items()}
        else:
            for gene in self.genes:
                gene_dfs[gene] = gene_dfs[gene].reindex(columns=self.feature_names[gene], fill_value=0)

        gene_dfs = {gene: FeatureScaler.normalize(df.values, self.normalization_type) for gene, df in gene_dfs.items()}
        examples = NumpyHelper.concat_arrays_rowwise([gene_dfs[gene] for gene in self.genes])

        if params.encode_labels:
            labels = dataset.get_metadata(params.label_config.get_labels_by_name(), return_df=False)
        else:
            labels = None

        return self._make_encoded_dataset(dataset, examples, labels, params)

    def _make_encoded_dataset(self, dataset, examples, labels, params: EncoderParams):

        examples = self._scale_examples(examples, params)

        encoded_data = EncodedData(examples=examples, labels=labels, example_ids=dataset.get_example_ids(),
                                   feature_names=self.all_feature_names, encoding=GeneFrequencyEncoder.__name__,
                                   info={'genes': self.genes},
                                   feature_annotations=pd.DataFrame({"feature": self.all_feature_names}))

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = encoded_data
        return encoded_dataset

    def _scale_examples(self, examples, params):
        if params.learn_model:
            self.scaler = StandardScaler(with_mean=self.scale_to_zero_mean, with_std=self.scale_to_unit_variance)
            examples = FeatureScaler.standard_scale_fit(self.scaler, examples, with_mean=self.scale_to_zero_mean)
        else:
            examples = FeatureScaler.standard_scale(self.scaler, examples, with_mean=self.scale_to_zero_mean)

        return examples



from pathlib import Path

import pandas as pd

from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.data_model.SequenceParams import RegionType
from scripts.specification_util import update_docs_per_mapping


class OLGAImport(DataImport):
    """
    Imports data generated by `OLGA <https://github.com/statbiophys/OLGA>`_ simulations into a Repertoire-, or SequenceDataset. Assumes that the columns in each
    file correspond to: nucleotide sequences, amino acid sequences, v genes, j genes

    Reference: Sethna, Zachary et al.
    ‘High-throughput immune repertoire analysis with IGoR’. Bioinformatics, (2019)
    `doi.org/10.1093/bioinformatics/btz035 <https://doi.org/10.1093/bioinformatics/btz035>`_.

    **Specification arguments:**

    - path (str): For RepertoireDatasets, this is the path to a directory with OLGA files to import. For Sequence- or ReceptorDatasets this path may either be the path to the file to import, or the path to the folder locating one or multiple files with .tsv, .csv or .txt extensions. By default path is set to the current working directory.

    - is_repertoire (bool): If True, this imports a RepertoireDataset. If False, it imports a SequenceDataset. By default, is_repertoire is set to True.

    - metadata_file (str): Required for RepertoireDatasets. This parameter specifies the path to the metadata file. This is a csv file with columns filename, subject_id and arbitrary other columns which can be used as labels in instructions. Only the OLGA files included under the column 'filename' are imported into the RepertoireDataset. SequenceDataset metadata is currently not supported.

    - import_illegal_characters (bool): Whether to import sequences that contain illegal characters, i.e., characters that do not appear in the sequence alphabet (amino acids including stop codon '*', or nucleotides). When set to false, filtering is only applied to the sequence type of interest (when running immuneML in amino acid mode, only entries with illegal characters in the amino acid sequence are removed). By default import_illegal_characters is False.

    - import_empty_nt_sequences (bool): imports sequences which have an empty nucleotide sequence field; can be True or False. By default, import_empty_nt_sequences is set to True.

    - import_empty_aa_sequences (bool): imports sequences which have an empty amino acid sequence field; can be True or False; for analysis on amino acid sequences, this parameter should be False (import only non-empty amino acid sequences). By default, import_empty_aa_sequences is set to False.

    - region_type (str): Which part of the sequence to import. By default, this value is set to IMGT_CDR3. This means the first and last amino acids are removed from the CDR3 sequence, as OLGA uses the IMGT junction. Specifying any other value will result in importing the sequences as they are. Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.

    - separator (str): Column separator, for OLGA this is by default "\\t".

    - column_mapping (dict): defines which columns to import from olga format: keys are the number of the columns and values are the names of the AIRR fields to be mapped to. For OLGA, this is by default set to:

        .. indent with spaces
        .. code-block:: yaml

            0: junction
            1: junction_aa
            2: v_call
            3: j_call


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            datasets:
                my_olga_dataset:
                    format: OLGA
                    params:
                        path: path/to/files/
                        is_repertoire: True # whether to import a RepertoireDataset (True) or a SequenceDataset (False)
                        metadata_file: path/to/metadata.csv # metadata file for RepertoireDataset
                        import_illegal_characters: False # remove sequences with illegal characters for the sequence_type being used
                        import_empty_nt_sequences: True # keep sequences even though the nucleotide sequence might be empty
                        import_empty_aa_sequences: False # filter out sequences if they don't have amino acid sequence set
                        # Optional fields with OLGA-specific defaults, only change when different behavior is required:
                        separator: "\\t" # column separator
                        columns_to_load: [0, 1, 2, 3]
                        column_mapping:
                            0: junction
                            1: junction_aa
                            2: v_call
                            3: j_call

    """

    def load_file(self, filename: Path) -> pd.DataFrame:
        df = pd.read_csv(str(filename), sep=self.params.separator, iterator=False, dtype=str, header=None,
                         usecols=list(self.params.column_mapping.keys()),
                         names=list(self.params.column_mapping.values()))

        if sorted(list(self.params.column_mapping.values())) == ['j_call', 'junction', 'junction_aa', 'v_call']\
                and not any(col in df.columns for col in ['cdr3', 'cdr3_aa']):

            df['cdr3'] = df['junction'].str[3:-3]
            df['cdr3_aa'] = df['junction_aa'].str[1:-1]

        return df

    @staticmethod
    def get_documentation():
        doc = str(OLGAImport.__doc__)

        region_type_values = str([region_type.name for region_type in RegionType])[1:-1].replace("'", "`")

        mapping = {
            "Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.": f"Valid values are {region_type_values}.",
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



import os
import shutil
from pathlib import Path

import yaml

from immuneML.api.galaxy.GalaxyTool import GalaxyTool
from immuneML.api.galaxy.Util import Util
from immuneML.app.ImmuneMLApp import ImmuneMLApp
from immuneML.util.Logger import print_log
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.ml_model_application.MLApplicationInstruction import MLApplicationInstruction


class GalaxyMLApplicationTool(GalaxyTool):

    def __init__(self, specification_path: Path, result_path: Path, **kwargs):
        Util.check_parameters(specification_path, result_path, kwargs, GalaxyMLApplicationTool.__name__)
        super().__init__(specification_path, result_path, **kwargs)

    def _run(self):
        PathBuilder.build(self.result_path)
        self._check_specs()
        state = ImmuneMLApp(self.yaml_path, self.result_path).run()[0]
        if os.path.relpath(state.predictions_path) != os.path.relpath(self.result_path / "predictions.csv"):
            shutil.copy(state.predictions_path, self.result_path / "predictions.csv")
        print_log("Applied ML model to the dataset, predictions are available.")

    def _check_specs(self):
        with open(self.yaml_path, "r") as file:
            specs = yaml.safe_load(file)

        instruction_name = Util.check_instruction_type(specs, GalaxyMLApplicationTool.__name__, MLApplicationInstruction.__name__[:-11])

        ParameterValidator.assert_keys_present(list(specs['instructions'][instruction_name].keys()), ["dataset", "config_path"],
                                               GalaxyMLApplicationTool.__name__, instruction_name)

        assert os.path.isfile(specs['instructions'][instruction_name]['config_path']), \
            f"{GalaxyMLApplicationTool.__name__}: file specified under 'config_path' parameter " \
            f"({specs['instructions'][instruction_name]['config_path']}) is not available. Please check if it was correctly uploaded or if the file" \
            f" name is correct."



import abc

from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.hyperparameter_optimization.HPSettingResult import HPSettingResult


class HPOptimizationStrategy(metaclass=abc.ABCMeta):
    """
    hyper-parameter optimization strategy is a base class of all different hyper-parameter optimization approaches,
    such as grid search, random search, bayesian optimization etc.

    HPOptimizationStrategy internally keeps a dict of settings that were tried out and the metric value that was
    obtained on the validation set which it then uses to determine the next step
    """

    def __init__(self, hp_settings: list, search_criterion=max):
        self.hp_settings = {hp_setting.get_key(): hp_setting for hp_setting in hp_settings}
        self.search_space_metric = {hp_setting.get_key(): None for hp_setting in hp_settings}
        self.search_criterion = search_criterion

    @abc.abstractmethod
    def generate_next_setting(self, hp_setting: HPSetting = None, metric: dict = None):
        """
        generator function which returns the next hyper-parameter setting to be evaluated
        :param hp_setting: previous setting (None if it's the first iteration)
        :param metric: performance metric from the previous setting per label
        :return: new hp_setting or None (if the end is reached)
        """
        pass

    @abc.abstractmethod
    def get_optimal_hps(self) -> HPSetting:
        pass

    @abc.abstractmethod
    def get_all_hps(self) -> HPSettingResult:
        pass

    @abc.abstractmethod
    def clone(self):
        pass



# quality: peripheral
import itertools
import logging

from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.SequenceSet import ReceptorSequence, Repertoire
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.PositionHelper import PositionHelper


class KmerHelper:
    @staticmethod
    def create_kmers_from_sequence(sequence: ReceptorSequence, k: int, sequence_type: SequenceType, overlap: bool = True):
        return KmerHelper.create_kmers_from_string(sequence.get_sequence(sequence_type), k, overlap)

    @staticmethod
    def create_kmers_from_string(sequence, k: int, overlap: bool = True):
        kmers = []
        step = 1 if overlap else k
        for i in range(0, len(sequence) - k + 1, step):
            kmers.append(sequence[i:i + k])
        return kmers

    @staticmethod
    def create_IMGT_kmers_from_sequence(sequence: ReceptorSequence, k: int, sequence_type: SequenceType, region_type: RegionType = RegionType.IMGT_CDR3):
        return KmerHelper.create_IMGT_kmers_from_string(sequence.get_sequence(sequence_type), k, region_type)

    @staticmethod
    def create_IMGT_kmers_from_string(sequence: str, k: int, region_type: RegionType):
        positions = PositionHelper.gen_imgt_positions_from_length(len(sequence), region_type)
        if positions is not None and len(positions) > 0:
            sequence_w_pos = list(zip(list(sequence), positions))
            kmers = KmerHelper.create_kmers_from_string(sequence_w_pos, k)
            kmers = [(''.join([x[0] for x in kmer]), kmer[0][1]) for kmer in kmers]
            return kmers
        else:
            logging.warning(f"{KmerHelper.__name__}: {sequence} could not be represented using IMGT {k}-mers, "
                            f"no IMGT positions were found. Returning empty list instead...")
            return []

    @staticmethod
    def create_IMGT_gapped_kmers_from_sequence(sequence: ReceptorSequence, sequence_type: SequenceType, k_left: int,
                                               max_gap: int, k_right: int = None, min_gap: int = 0,
                                               region_type: RegionType = RegionType.IMGT_CDR3):
        positions = PositionHelper.gen_imgt_positions_from_sequence(sequence, sequence_type, region_type)

        sequence_w_pos = list(zip(list(sequence.get_sequence(sequence_type)), positions))
        kmers = KmerHelper.create_gapped_kmers_from_string(sequence_w_pos, k_left=k_left, max_gap=max_gap,
                                                           k_right=k_right, min_gap=min_gap)
        if kmers is not None:
            kmers = [(''.join([x[0] for x in kmer]), kmer[0][1]) for kmer in kmers]
            return kmers
        else:
            return None

    @staticmethod
    def create_gapped_kmers_from_string(sequence, k_left: int, max_gap: int, k_right: int = None, min_gap: int = 0):
        length = len(sequence)
        k_right = k_left if k_right is None else k_right

        if length < k_left + k_right + max_gap:
            raise ValueError('Sequence length is less than k_left + k_right + max_gap. '
                             'Filter sequences from each repertoire that are less than this length then rerun.')
        gapped_kmers = []
        for i in range(min_gap, max_gap + 1):
            s = k_left + k_right + i
            kmers = [sequence[i: i + s] for i in range(length - s + 1)]
            if isinstance(sequence, str):
                gapped_kmers.extend([kmer[:k_left] + i * "." + kmer[k_left + i:] for kmer in kmers])
            if isinstance(sequence, list):
                gapped_kmers.extend([kmer[:k_left] + [(".", el[1]) for el in kmer[k_left:k_left+i]] + kmer[k_left + i:] for kmer in kmers])
        return gapped_kmers

    @staticmethod
    def create_gapped_kmers_from_sequence(sequence: ReceptorSequence, sequence_type: SequenceType, k_left: int, max_gap: int, k_right: int = None,
                                          min_gap: int = 0):
        return KmerHelper.create_gapped_kmers_from_string(sequence.get_sequence(sequence_type), k_left, max_gap, k_right, min_gap)

    @staticmethod
    def create_all_kmers(k: int, alphabet: list):
        """
        creates all possible k-mers given a k-mer length and an alphabet
        :param k: length of k-mer (int)
        :param alphabet: list of characters from which to make all possible k-mers (list)
        :return: alphabetically sorted list of k-mers
        """
        kmers = [''.join(x) for x in itertools.product(alphabet, repeat=k)]
        kmers.sort()
        return kmers

    @staticmethod
    def create_sentences_from_repertoire(repertoire: Repertoire, k: int, sequence_type: SequenceType, overlap: bool = True, region_type: RegionType = RegionType.IMGT_CDR3):
        sentences = []
        for sequence in repertoire.sequences(region_type):
            sentences.append(KmerHelper.create_kmers_from_sequence(sequence=sequence, k=k, overlap=overlap, sequence_type=sequence_type))
        return sentences

    @staticmethod
    def create_kmers_within_HD(kmer: str, alphabet: list, distance: int = 1):

        assert distance < len(kmer)

        if distance > 1:
            logging.warning("In create_kmers_within_HD distance larger than 1 is not yet implemented. "
                          "Using default value 1...", Warning)

        pairs = []

        for i in range(len(kmer)):
            for letter in alphabet:
                new_kmer = kmer[0:i] + letter + kmer[i + 1:]
                pairs.append([kmer, new_kmer])

        return pairs



from sklearn.cluster import KMeans as SklearnKmeans

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.ml_methods.clustering.ClusteringMethod import ClusteringMethod, get_data_for_clustering


class KMeans(ClusteringMethod):
    """
    k-means clustering method which wraps scikit-learn's KMeans. Input arguments for the method are the
    same as supported by scikit-learn (see `KMeans scikit-learn documentation
    <https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html>`_ for details).

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_kmeans:
                    KMeans:
                        # arguments as defined by scikit-learn
                        n_clusters: 2
    """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)
        self.model = SklearnKmeans(**kwargs)

    def fit(self, dataset: Dataset):
        data = get_data_for_clustering(dataset)
        self.model.fit(data)

    def fit_predict(self, dataset: Dataset):
        data = get_data_for_clustering(dataset)
        return self.model.fit_predict(data)

    def predict(self, dataset: Dataset):
        data = get_data_for_clustering(dataset)
        return self.model.predict(data)

    def transform(self, dataset: Dataset):
        data = get_data_for_clustering(dataset)
        return self.model.transform(data)



import pandas as pd

from immuneML.data_model.datasets.ElementDataset import ReceptorDataset, ElementDataset, SequenceDataset
from immuneML.data_model.EncodedData import EncodedData
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.util.EncoderHelper import EncoderHelper
from immuneML.util.Logger import print_log


class TCRdistEncoder(DatasetEncoder):
    """
    Encodes the given ReceptorDataset as a distance matrix between all receptors, where the distance is computed using TCRdist from the paper:
    Dash P, Fiore-Gartland AJ, Hertz T, et al. Quantifiable predictive features define epitope-specific T cell receptor repertoires.
    Nature. 2017; 547(7661):89-93. `doi:10.1038/nature22383 <https://www.nature.com/articles/nature22383>`_.

    For the implementation, `TCRdist3 <https://tcrdist3.readthedocs.io/en/latest/>`_ library was used (source code available
    `here <https://github.com/kmayerb/tcrdist3>`_).

    **Dataset type:**

    - ReceptorDataset

    - SequenceDataset


    **Specification arguments:**

    - cores (int): number of processes to use for the computation

    - cdr3_only (bool): whether to use only cdr3 or also v gene; if set to false, encoding will only compute the distances
      between the CDR3 regions of the receptors


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_tcr_dist_enc:
                    TCRdist:
                        cores: 4
                        cdr3_only: false # default tcrdist behavior

    """

    def __init__(self, cores: int, cdr3_only: bool, name: str = None):
        super().__init__(name=name)
        self.cores = cores
        self.cdr3_only = cdr3_only
        self.distance_matrix = None
        self.context = None
        self._tmp_results_path = None

    @staticmethod
    def build_object(dataset, **params):
        if isinstance(dataset, ReceptorDataset) or isinstance(dataset, SequenceDataset):
            return TCRdistEncoder(**params)
        else:
            raise ValueError("TCRdistEncoder is defined for receptor and sequence dataset.")

    def set_context(self, context: dict):
        self.context = context
        return self

    def encode(self, dataset, params: EncoderParams):
        if self._tmp_results_path is None and params.learn_model:
            self._tmp_results_path = params.result_path

        train_receptor_ids = EncoderHelper.prepare_training_ids(dataset, params, self._tmp_results_path)

        if params.learn_model:
            self._build_tcr_dist_matrix(dataset, params.label_config.get_labels_by_name())

        distance_matrix = self.distance_matrix.loc[dataset.get_example_ids(), train_receptor_ids]
        labels = self._build_labels(dataset, params) if params.encode_labels else None

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(examples=distance_matrix.values, labels=labels, example_ids=distance_matrix.index.values,
                                                   encoding=TCRdistEncoder.__name__)
        return encoded_dataset

    def _build_tcr_dist_matrix(self, dataset: ElementDataset, label_names):
        from immuneML.util.TCRdistHelper import TCRdistHelper

        chains = TCRdistHelper.get_chains(dataset)

        current_dataset = dataset if self.context is None or "dataset" not in self.context else self.context["dataset"]
        tcr_rep = TCRdistHelper.compute_tcr_dist(current_dataset, label_names, self.cores,self.cdr3_only)

        data = 0.

        if 'alpha' in chains:
            data += tcr_rep.pw_alpha
        if 'beta' in chains:
            data += tcr_rep.pw_beta

        self.distance_matrix = pd.DataFrame(data, index=tcr_rep.clone_df.clone_id.values,
                                            columns=tcr_rep.clone_df.clone_id.values)

    def _build_labels(self, dataset: ElementDataset, params: EncoderParams) -> dict:
        return dataset.get_metadata(params.label_config.get_labels_by_name())




from immuneML.util.CompAIRRHelper import CompAIRRHelper


class CompAIRRBatchIterator:
    def __init__(self, paths, sequence_batch_size):
        self.repertoire_ids = None
        self.sequence_batch_size = sequence_batch_size
        self.batch_paths = self.get_batch_dict(paths)
        self.sequence_count = self.compute_sequence_count()

    def __iter__(self):
        return self.get_sequence_vectors(self.repertoire_ids)

    def __len__(self):
        return self.sequence_count

    def compute_sequence_count(self):
        sequence_count = (len(self.batch_paths) - 1) * self.sequence_batch_size

        last_batch_path = self.batch_paths[max(self.batch_paths.keys())]
        sequence_count += len(CompAIRRHelper.read_compairr_output_file(last_batch_path))

        return sequence_count

    def get_batch_dict(self, paths):
        return {self.get_batch_from_path(path): path for path in paths}

    def get_batch_from_path(self, path):
        return int(path.stem.split("_batch")[1])

    def set_repertoire_ids(self, repertoire_ids):
        self.repertoire_ids = repertoire_ids

    def get_batches(self, repertoire_ids=None):
        for batch_idx in sorted(self.batch_paths):
            path = self.batch_paths[batch_idx]
            batch = CompAIRRHelper.read_compairr_output_file(path)

            if repertoire_ids is not None:
                batch = batch[repertoire_ids]

            # count clones only
            batch[batch > 1] = 1
            batch.sort_index(inplace=True)

            yield batch

    def get_sequence_vectors(self, repertoire_ids=None):
        for batch in self.get_batches(repertoire_ids):
            for idx, sequence_vector in batch.iterrows():
                yield sequence_vector.to_numpy()



from dataclasses import dataclass
from pathlib import Path

from immuneML.data_model.SequenceParams import ChainPair, RegionType
from immuneML.util.ParameterValidator import ParameterValidator


@dataclass
class DatasetImportParams:
    path: Path = None
    is_repertoire: bool = None
    metadata_file: Path = None
    paired: bool = None
    receptor_chains: ChainPair = None
    result_path: Path = None
    columns_to_load: list = None
    label_columns: list = None
    separator: str = None
    column_mapping: dict = None
    column_mapping_synonyms: dict = None
    region_type: RegionType = None
    import_productive: bool = None
    import_unknown_productivity: bool = None
    import_unproductive: bool = None
    import_with_stop_codon: bool = None
    import_out_of_frame: bool = None
    import_illegal_characters: bool = None
    number_of_processes: int = 1
    sequence_file_size: int = 50000
    organism: str = None
    import_empty_nt_sequences: bool = None
    import_empty_aa_sequences: bool = None
    dataset_file: Path = None

    @classmethod
    def build_object(cls, path: Path = None, metadata_file: Path = None, result_path: Path = None,
                     region_type: str = None, receptor_chains: str = None, **kwargs):
        params = {
            "path": Path(path) if path is not None else None,
            "metadata_file": Path(metadata_file) if metadata_file is not None else None,
            "result_path": Path(result_path) if result_path is not None else None,
            "region_type": RegionType[region_type.upper()] if region_type else None,
            "receptor_chains": ChainPair[receptor_chains.upper()] if receptor_chains else None,
        }

        if "column_mapping" in kwargs and kwargs['column_mapping']:
            ParameterValidator.assert_type_and_value(kwargs['column_mapping'], dict, cls.__name__, 'column_mapping')

            if not all(isinstance(el, int) for el in kwargs['column_mapping'].keys()):
                assert len(set(kwargs['column_mapping'].values())) == len(list(kwargs['column_mapping'].values())), \
                    (f"{cls.__name__}: Columns must be mapped to unique names, got: "
                     f"{list(kwargs['column_mapping'].values())}.")

        if kwargs.get('columns_to_load', None):
            ParameterValidator.assert_type_and_value(kwargs['columns_to_load'], list, cls.__name__, "columns_to_load")

        if kwargs.get('label_columns', None):
            ParameterValidator.assert_type_and_value(kwargs['label_columns'], list, cls.__name__, "label_columns")

        if kwargs.get('columns_to_load', None) and kwargs.get('label_columns', None):
            assert all(col_name in kwargs['columns_to_load'] for col_name in kwargs['label_columns']), \
                (f"{cls.__name__}: Some column names defined under 'label_columns' were not listed in 'columns_to_load'.\n"
                 f"label_columns: {kwargs['label_columns']}\n"
                 f"columns_to_load: {kwargs['columns_to_load']}\n"
                 f"To prevent this error, please add all label columns to columns_to_load.")

        if kwargs.get('columns_to_load', None) and kwargs.get('column_mapping', None):
            assert all(key in kwargs['columns_to_load'] for key in kwargs['column_mapping']), \
                f"{cls.__name__}: Some keys defined under 'column_mapping' were not listed in 'columns_to_load'."

            total_specified_fields = list(kwargs['column_mapping'].keys()) + kwargs['columns_to_load']
            if kwargs.get('label_columns', None):
                total_specified_fields += kwargs['label_columns']

            total_specified_fields = set(total_specified_fields)

            number_specified_fields = len(total_specified_fields)
            number_imported_fields = len(kwargs['columns_to_load'])
            assert number_specified_fields == number_imported_fields, \
                (f"{cls.__name__}: 'column_mapping', 'columns_to_load' and 'label_columns' fields were not correctly specified: "
                 f"{total_specified_fields} fields specified, and {number_imported_fields} fields are imported.\n"
                 f"All specified fields are: {sorted(total_specified_fields)}\n"
                 f"This does not match columns_to_load: {sorted(kwargs['columns_to_load'])}")

        params = {**kwargs, **params}
        return DatasetImportParams(**params)
